# SPDX-License-Identifier: Apache-2.0
import pandas
from ._supersample_utility import *
from ._supersample_utility import _find_rotation_


DEFAULT_COLUMNS = ["x", "y"]
DEFAULT_WITH_DEPTH = ["x", "depth", "y"]
TMP_INDEX = "__index__"


def per_pixel_coordinate_transformation(fm, orient, from_system="global", to_system="rotated"):
    """
    Systems:
    global: The global coordinate system in um or voxel indices.
    localized: The global coordinate system, but origin moved to the pixel center
    rotated: A local coordinate system, origin at the pixel center, y-axis rotated vertical
    rotated_flat: A local coordinate system, origin at the pixel center, y-axis rotated vertical and then flattened away
    through parallel projection
    subpixel: A local coordinate system, origin at the pixel center, y-axis rotated vertical and then flattened away
    through parallel projection, x and z axes oriented like the flat-x and flat-y axes of the flat map.
    subpixel_depth: A local coordinate system, origin at the pixel center, y-axis rotated vertical,
    x and z axes oriented like the flat-x and flat-y axes of the flat map.
    """
    lst_systems = ["global", "localized", "rotated", "rotated_flat", "subpixel", "subpixel_depth"]
    try:
        tgt_tf = (lst_systems.index(from_system), lst_systems.index(to_system))
    except ValueError:
        raise ValueError("from_system and to_system must be in: {0}, but you provided {1}".format(lst_systems,
                                                                                                  (from_system,
                                                                                                   to_system)))
    invalid_combos = [(2, 3), (3, 2), (3, 1), (3, 0), (4, 2), (4, 1), (4, 0), (3, 5), (4, 5), (5, 3), (5, 4)]
    if tgt_tf in invalid_combos:
        raise ValueError("Invalid combination!")
    if tgt_tf[0] == tgt_tf[1]:
        raise ValueError("Identity transformation not supported")

    vxl_frame = voxel_flat_coordinate_frame(fm)
    per_pixel = vxl_frame.groupby(["f_x", "f_y"])

    per_pixel_negative_center = per_pixel.apply(lambda x: -numpy.mean(x.values, axis=0))
    global2localized = per_pixel_negative_center.apply(Translation)
    if tgt_tf == (0, 1): return global2localized
    if tgt_tf == (1, 0): return global2localized.apply(lambda x: x.inv())

    per_pixel_orient = per_pixel_negative_center.apply(lambda x: orient.lookup(-x))
    localized2rotated = per_pixel_orient.apply(lambda o_vec: Rotation.from_quat(numpy.hstack([o_vec[1:],
                                                                                       o_vec[0:1]])).inv())
    if tgt_tf == (1, 2): return localized2rotated
    if tgt_tf == (2, 1): return localized2rotated.apply(lambda x: x.inv())

    global2rotated = global2localized.combine(localized2rotated, Combination)
    if tgt_tf == (0, 2): return global2rotated
    if tgt_tf == (2, 0): return global2rotated.apply(lambda x: x.inv())

    tf_to_local_flat = Projection([0, 2])
    global2rotflat = global2rotated.apply(lambda base_tf: Combination(base_tf, tf_to_local_flat))
    if tgt_tf == (0, 3): return global2rotflat
    if tgt_tf == (3, 0): return global2rotflat.apply(lambda x: x.inv())

    localized2rotflat = localized2rotated.apply(lambda base_tf: Combination(base_tf, tf_to_local_flat))
    if tgt_tf == (1, 3): return localized2rotflat

    dfx, dfy = flatmap_pixel_gradient(vxl_frame)
    dfx_frame = per_pixel_negative_center.index.to_frame().apply(lambda x: dfx[x["f_x"], x["f_y"]].reshape((1, -1)),
                                                            axis=1)
    dfy_frame = per_pixel_negative_center.index.to_frame().apply(lambda x: dfy[x["f_x"], x["f_y"]].reshape((1, -1)),
                                                            axis=1)
    # Above gradient vectors are in "localized" space. Convert to rotated_flat
    dfx_frame = localized2rotflat.combine(dfx_frame, lambda a, b: a.apply(b))
    dfy_frame = localized2rotflat.combine(dfy_frame, lambda a, b: a.apply(b))

    # We now know the directions of neighboring pixel in the rotated_flat systems
    # Figure out a rotation that transforms the direction vectors to the x or y-axes
    rotflat2pixel_err = dfx_frame.combine(dfy_frame, _find_rotation_)
    rotflat2pixel = rotflat2pixel_err.apply(lambda x: x[0])
    err = rotflat2pixel_err.apply(lambda x: x[1])
    print("Rotation errors: min: {0}, median: {1}, mean: {2}, std: {3}, max: {4}".format(
        err.min(), err.median(), err.mean(), err.std(), err.max()
    ))
    if tgt_tf == (3, 4): return rotflat2pixel
    if tgt_tf == (4, 3): return rotflat2pixel.apply(lambda x: x.inv())
    if tgt_tf == (2, 4): return rotflat2pixel.apply(lambda base_tf: Combination(tf_to_local_flat, base_tf))
    if tgt_tf == (1, 4): return localized2rotflat.combine(rotflat2pixel, Combination)
    if tgt_tf == (0, 4): return global2rotflat.combine(rotflat2pixel, Combination)

    rot2pixel = rotflat2pixel.apply(lambda x: x.expand())

    #  Turn the y-location relative to the pixel geometrical center into a proper depth.
    pp_rotated = per_pixel.apply(lambda x: x.values).combine(global2rotated, func=lambda a, b: b.apply(a))
    tl_to_depth = pp_rotated.apply(lambda locs: Translation(numpy.array([0, -numpy.max(locs[:, 1]), 0])))
    rot2pixel = rot2pixel.combine(tl_to_depth, Combination)

    if tgt_tf == (2, 5): return rot2pixel
    if tgt_tf == (5, 2): return rot2pixel.apply(lambda x: x.inv())
    if tgt_tf == (1, 5): return localized2rotated.combine(rot2pixel, Combination)
    if tgt_tf == (5, 1): return rot2pixel.combine(localized2rotated, lambda a, b: Combination(a.inv(), b.inv()))
    if tgt_tf == (0, 5): return global2rotated.combine(rot2pixel, Combination)
    if tgt_tf == (5, 0): return rot2pixel.combine(global2rotated, lambda a, b: Combination(a.inv(), b.inv()))
    raise ValueError("This should never happen!")


def estimate_flatmap_pixel_size(fm, orient):
    raise NotImplementedError()  # TODO: Implement


def supersample_flatmap(fm, orient, pixel_sz=34.0, include_depth=False):
    import voxcell
    to_system = "subpixel"
    shape = fm.raw.shape[:3] + (2,)
    if include_depth:
        to_system = "subpixel_depth"
        shape = fm.raw.shape[:3] + (3,)
    vxl_frame = voxel_flat_coordinate_frame(fm, grouped=True)
    vxl_index_frame = voxel_flat_coordinate_frame(fm, in_voxel_indices=True, grouped=True)
    tf = per_pixel_coordinate_transformation(fm, orient, to_system=to_system)
    subpixel_loc = vxl_frame.combine(tf, lambda a, b: b.apply(a))
    if include_depth:
        final_loc = subpixel_loc.index.to_series().combine(subpixel_loc,
                                                           lambda a, b: pixel_sz * numpy.array([a[0], 0, a[1]]) + b)
    else:
        final_loc = subpixel_loc.index.to_series().combine(subpixel_loc,
                                                           lambda a, b: pixel_sz * numpy.array(a) + b)
    final_loc_arr = numpy.vstack(final_loc.values)
    if include_depth:
        final_loc_arr[:, 1] = final_loc_arr[:, 1] * -1
    vxl_loc = numpy.vstack(vxl_index_frame.values)
    out_raw = -numpy.ones(shape, dtype=float)
    out_raw[vxl_loc[:, 0], vxl_loc[:, 1], vxl_loc[:, 2]] = final_loc_arr
    return voxcell.VoxelData(out_raw, fm.voxel_dimensions, offset=fm.offset)


def supersampled_locations(df_in, columns_xyz, columns_uvw=None, columns_out=None,
                           circ=None, fm=None, orient=None, pixel_sz=34.0,  #TODO: use DEFAULT_PIXEL_SZ
                           column_index=None, include_depth=False):
    from ..circuit_models.neuron_groups.sonata_extensions import load_atlas_data
    if circ is None:
        assert fm is not None and orient is not None, "Must provide circuit or flatmap and orientation atlas!"
    if fm is None:
        fm = load_atlas_data(circ, "flatmap")
    if orient is None:
        orient = load_atlas_data(circ, "orientation")
    if pixel_sz is None:
        pixel_sz = estimate_flatmap_pixel_size(fm, orient)
    if column_index is None:
        index_frame = df_in.index.to_frame(name=TMP_INDEX)
        df_in = pandas.concat([df_in, index_frame], axis=1)
        column_index = TMP_INDEX
    if columns_out is None:
        if include_depth:
            columns_out = DEFAULT_WITH_DEPTH
        else:
            columns_out = DEFAULT_COLUMNS
    to_system = "subpixel"
    if include_depth:
        to_system = "subpixel_depth"

    loc_frame, df_with_midx = pandas_flat_coordinate_frame(df_in, fm,
                                                           columns_xyz=columns_xyz, columns_uvw=columns_uvw,
                                                           grouped=True)
    tf = per_pixel_coordinate_transformation(fm, orient, to_system=to_system)
    idxx = loc_frame.index.intersection(tf.index)

    res = tf[idxx].combine(loc_frame[idxx], lambda a, b: a.apply(b))
    if include_depth:
        final = res.index.to_series().combine(res, lambda a, b: numpy.array([a[0], 0, a[1]]) * pixel_sz + b)
    else:
        final = res.index.to_series().combine(res, lambda a, b: numpy.array(a) * pixel_sz + b)
    final_frame = numpy.vstack(final.values)
    if include_depth:
        final_frame[:, 1] = final_frame[:, 1] * -1

    index = df_with_midx[column_index][idxx].values
    out = pandas.DataFrame(final_frame,
                           columns=columns_out,
                           index=index)  # TODO: check that this index is in the right order
    return out



# SPDX-License-Identifier: Apache-2.0
import os, pandas

from .defaults import VIRTUAL_FIBERS_FN

SONATA_CONNECTOME_FN = "edges.sonata"
STR_ATLAS_DIR = "atlas_dir"
ATLAS_LOC = "atlas"
HIERARCHY_FN = "hierarchy.json"
STR_NODE_SETS = "node_sets_file"


def circuit_base_dir(circ):
    if STR_NODE_SETS in circ.config:
        sonata_base_dir = os.path.split(circ.config[STR_NODE_SETS])[0]
        base_dir, sonata = os.path.split(sonata_base_dir)
        if sonata == "sonata":
            return base_dir
        return sonata_base_dir
    return None


def atlas_dir(circ):
    if STR_ATLAS_DIR in circ.config["components"].get("provenance", {}):
        return circ.config["components"]["provenance"][STR_ATLAS_DIR]
    else:
        circ_base = circuit_base_dir(circ)
        if circ_base is None: return None
        atlas = os.path.join(circ_base, ATLAS_LOC)
        if os.path.exists(atlas): return atlas
    return None

def load_atlas_data(circ, atlas_name):
    import voxcell

    atlas = atlas_dir(circ)
    if atlas is None: raise RuntimeError("No atlas directory found!")
    atlas_fn = os.path.join(atlas, atlas_name)
    if os.path.splitext(atlas_fn)[1] != ".nrrd":
        atlas_fn = atlas_fn + ".nrrd"
    
    return voxcell.VoxelData.load_nrrd(atlas_fn)

def load_atlas_hierarchy(circ):
    import voxcell

    atlas = atlas_dir(circ)
    if atlas is None: raise RuntimeError("No atlas directory found!")
    hier_fn = os.path.join(atlas, HIERARCHY_FN)
    return voxcell.RegionMap.load_json(hier_fn)


def simulation_conditions(sim):
    """
    Return simulation conditions of a bluepysnap.Simulation, looked up from
    the associated sim_conditions.json, if it exists
    """
    from os import path
    import json
    
    sim_path = sim._config._config_dir
    sim_json = sim_path + "sim_conditions.json"
    if not path.isfile(sim_json):
        return {}
    with open(sim_json, "r") as fid:
        conds = json.load(fid)
    return conds



# SPDX-License-Identifier: Apache-2.0
"""
Contributed by Vishal Sood
Last changed: 2021/11/29
"""
import os
from pathlib import Path
import logging, logging.config

logging.basicConfig(level=logging.NOTSET)
ERROR_FORMAT = ("%(levelname)s at %(asctime)s in %(funcName)s in %(pathname) "
                "at line %(lineno)d: %(message)s")
INFO_FORMAT = " %(asctime)s: %(message)s"
DEBUG_FORMAT = "%(lineno)d in %(pathname)s at %(asctime)s: %(message)s"
LOG_CONFIG = {'version':1,
                'formatters':{'error':{'format': ERROR_FORMAT},
                              "info": {"format": INFO_FORMAT},
                              'debug':{'format': DEBUG_FORMAT}},
                'handlers':{'console':{'class': 'logging.StreamHandler',
                                        'formatter': 'info',
                                        'level': logging.INFO},
                            'file':{'class': 'logging.FileHandler',
                                    'filename': Path.cwd()/"topology_analysis.log",
                                    'formatter': 'error',
                                    'level': logging.ERROR}},
                'root':{'handlers':('console', 'file')}}

logging.config.dictConfig(LOG_CONFIG)


def get_logger(for_step, at_level=None):
    """Centralized logging for ConnSense Apps."""
    logger = logging.getLogger(f"ConnSense {for_step.upper()}")
    logger.setLevel(at_level if at_level
                    else os.environ.get("LOGLEVEL", logging.INFO))
    return logger



# SPDX-License-Identifier: Apache-2.0
from . import neuron_groups
from .connection_matrix import connection_matrix_for_gids, connection_matrix_between_groups_partial, connection_matrix_between_groups_partition
from .connection_matrix import circuit_connection_matrix, circuit_matrix_between_groups, circuit_group_matrices
from .connection_matrix import full_connection_matrix, circuit_cross_group_matrices, circuit_node_set_matrix
from .input_spikes import input_innervation, input_innervation_from_matrix
from .neuron_groups.sonata_extensions import simulation_conditions



# SPDX-License-Identifier: Apache-2.0
import io
import h5py
import os
import numpy
import pandas as pd


def write_sparse_matrix_payload(hdf_group, dset_pattern="matrix_{0}"):
    def write_dset(mat):
        dset_name = dset_pattern.format(len(hdf_group.keys()))
        from scipy import sparse
        bio = io.BytesIO()
        sparse.save_npz(bio, mat)
        bio.seek(0)
        matrix_bytes = list(bio.read())
        hdf_group.create_dataset(dset_name, data=matrix_bytes)
        return hdf_group.name + "/" + dset_name
    return write_dset


def write_dense_matrix_payload(hdf_group, dset_pattern="matrix_{0}"):
    def write_dset(mat):
        dset_name = dset_pattern.format(len(hdf_group.keys()))
        hdf_group.create_dataset(dset_name, data=mat)
        return hdf_group.name + "/" + dset_name
    return write_dset


def read_sparse_matrix_payload(hdf_dset):
    from scipy import sparse
    raw_data = bytes(hdf_dset[:].astype(numpy.uint8))
    bio = io.BytesIO(raw_data)
    mat = sparse.load_npz(bio)
    return mat


def read_dense_matrix_payload(hdf_dset):
    return numpy.array(hdf_dset)


READ_WRITE = {
    "sparse": (read_sparse_matrix_payload, write_sparse_matrix_payload),
    "dense": (read_dense_matrix_payload, write_dense_matrix_payload)
}


def write_toc_plus_payload(extracted, to_path, payload_type="sparse", format=None):
    # TODO: Could inspect the contents of extraced and determine the payload_type from it.
    path_hdf_store, group_identifier = to_path
    group_identifier_toc = group_identifier + "/toc"
    group_identifier_mat = group_identifier + "/payload"
    func = READ_WRITE[payload_type][1]

    h5_file = h5py.File(path_hdf_store, "a")
    h5_grp_mat = h5_file.require_group(group_identifier_mat)
    h5_grp_mat.attrs["payload_type"] = payload_type
    toc = extracted.apply(func(h5_grp_mat))
    h5_file.close()

    toc.to_hdf(path_hdf_store, key=group_identifier_toc,
                     mode="a", format=(format or "fixed"))


class LazyMatrix:
    """..."""
    from lazy import lazy

    def __init__(self, path_hdf, path_dset, func):
        """..."""
        self._hdf = path_hdf
        self._dset = path_dset
        self._reader = func

    @lazy
    def matrix(self):
        """..."""
        with h5py.File(self._hdf, 'r') as hdf:
            dset = hdf[self._dset]
            matrix = self._reader(dset)
        return matrix


def read_toc_plus_payload(path):
    path_hdf_store, group_identifier = path
    group_identifier_toc = group_identifier + "/toc"
    group_identifier_mat = group_identifier + "/payload"
    with h5py.File(path_hdf_store, "r") as h5:
        func = READ_WRITE[h5[group_identifier_mat].attrs.get("payload_type", "sparse")]  # Default for bw compatibility

    if not os.path.isfile(path_hdf_store):
        raise RuntimeError(f"Missing HDF data at path {path_hdf_store}\n")

    toc = pd.read_hdf(path_hdf_store, key=group_identifier_toc)

    return toc.apply(lambda dset_path: LazyMatrix(path_hdf_store, dset_path, func))


# SPDX-License-Identifier: Apache-2.0
import numpy

from .supersampling import supersample_flatmap


def _loader(recipe):
    if isinstance(recipe, str):
        import yaml
        with open(recipe, "r") as fid:
            mp = yaml.load(fid, Loader=yaml.SafeLoader)
        return mp
    return recipe


def regions_of_population(nm, recipe):
    tmp = [x for x in recipe["populations"] if x["name"] == nm]
    assert len(tmp) == 1
    regions = tmp[0]['atlas_region']
    if isinstance(regions, list):
        return [region["name"] for region in regions]
    return [regions['name']]


def ids_of_regions(names_region, hier):
    reg_ids = set()
    for region in names_region:
        reg_ids.update(hier.find(region, "acronym", with_descendants=True))
    return list(reg_ids)


def twod2rgb(flat_coords, x, y):
    pts = flat_coords - numpy.array([[x[2], y[2]]])
    T = numpy.array([[x[0] - x[2], x[1] - x[2]],
                    [y[0] - y[2], y[1] - y[2]]])
    Tinv = numpy.linalg.inv(T)
    l_0_1 = numpy.dot(Tinv, pts.transpose())
    l = numpy.vstack([l_0_1, 1.0 - numpy.sum(l_0_1, axis=0, keepdims=True)]).transpose()
    l[l < 0] = 0.0; l[l > 1] = 1.0
    return l


def twod2mapping_coords(flat_coords, x, y):
    pts = flat_coords - numpy.array([[x[2], y[2]]])
    v0 = numpy.array([x[0], y[0]]) - numpy.array([x[2], y[2]])
    v1 = numpy.array([x[1], y[1]]) - numpy.array([x[2], y[2]])
    v0 = v0 / numpy.linalg.norm(v0); v1 = v1 / numpy.linalg.norm(v1)
    x_out = numpy.dot(pts, v0.reshape((2, 1)))  # N x 2 * 2 x 1
    y_out = numpy.dot(pts, v1.reshape((2, 1)))
    return numpy.hstack([x_out, y_out])


def atlas_of_mapping_coordinates(recipe, circ=None, fm=None, orient=None, hier=None, ann=None,
                                 supersample=False):
    import voxcell
    from ..circuit_models.neuron_groups.sonata_extensions import load_atlas_data, load_atlas_hierarchy

    if fm is None or orient is None or hier is None or ann is None:
        assert circ is not None, "Must provide all of fm, orient, hier and ann or a Circuit!"
        if fm is None:
            fm = load_atlas_data(circ, "flatmap")
        if orient is None:
            orient = load_atlas_data(circ, "orientation")
        if hier is None:
            hier = load_atlas_hierarchy(circ)
        if ann is None:
            ann = load_atlas_data(circ, "brain_regions")
    elif circ is not None:
        print("Provided Circuit will be ignored since all of fm, orient, hier and ann were provided")
    
    if supersample:
        fm = supersample_flatmap(fm, orient, pixel_sz=1.0)
    recipe = _loader(recipe)

    touched_srcs = []
    out_raw = -numpy.ones_like(fm.raw, dtype=int)

    for proj in recipe["projections"]:
        src_str = proj["source"].split("_")[0]
        if src_str in touched_srcs:
            continue
        touched_srcs.append(src_str)

        pop_strs = regions_of_population(proj["source"], recipe)
        pop_ids = ids_of_regions(pop_strs, hier)
        pop_mask = numpy.in1d(ann.raw.flat, pop_ids).reshape(ann.raw.shape)
        pop_flat_coords = fm.raw[pop_mask]

        x = numpy.array(proj["mapping_coordinate_system"]["x"])
        y = numpy.array(proj["mapping_coordinate_system"]["y"])

        out_raw[pop_mask] = twod2mapping_coords(pop_flat_coords, x, y)
    
    return voxcell.VoxelData(out_raw, fm.voxel_dimensions, offset=fm.offset)






# SPDX-License-Identifier: Apache-2.0
from numpy.random.mtrand import RandomState
import pandas
import json
import os

from numpy import random
from scipy import sparse


from conntility import analysis as test_module
from conntility import ConnectivityMatrix


TEST_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_DATA_DIR = os.path.join(TEST_DIR, "data")

RANDOM_WORDS = ["banana", "apple", "orange", "tangerine", "car"]

sz = 500
nrn = pandas.DataFrame(
    random.randint(0, 5, size=(sz, 3)),
    columns=["mtype", "etype", "ttype"]
)
nrn["mtype"] = [RANDOM_WORDS[i] for i in nrn["mtype"]]
M = sparse.csc_matrix(random.rand(sz, sz) < 0.06)

with open(os.path.join(TEST_DATA_DIR, "test_analysis_config.json"), "r") as fid:
    ana_cfg_raw = json.load(fid)
with open(os.path.join(TEST_DATA_DIR, "test_analysis_control_sample.json"), "r") as fid:
    ana_cfg_sample = json.load(fid)

def test_analysis_by_config():
    ana_cfg = ana_cfg_raw.copy()
    rnd_er = ana_cfg["analyses"]["simplex_counts"]["source"] =\
        os.path.join(TEST_DATA_DIR, ana_cfg["analyses"]["simplex_counts"]["source"])
    rnd_er = ana_cfg["analyses"]["simplex_counts"]["decorators"][1]["analysis_arg"]["random_er"]
    rnd_er["source"] = os.path.join(TEST_DATA_DIR, rnd_er["source"])

    ana_cfg["analyses"]["random_number"]["source"] =\
        os.path.join(TEST_DATA_DIR, ana_cfg["analyses"]["random_number"]["source"])

    analyses = list(test_module.get_analyses(ana_cfg))

    for analysis in analyses:
        if analysis.name == "random_number":
            result = analysis.apply(M, nrn)
            assert len(result.index.names) == 2
            assert tuple(result.index.names) == ("idx-mtype", "idx-etype")
        elif analysis.name == "simplex_counts":
            result = analysis.apply(M, nrn).unstack("Control")
            nrm_result = (result["data"] - result["random_er"]) / result["random_er"]

            assert (nrm_result[:, 0] == 0).all()
            for word in RANDOM_WORDS:
                assert word in nrm_result
            assert nrm_result["orange"].index.name == "dim"


def test_manual_filtering_config():
    decorator = test_module.analysis_decorators.grouped_by_filtering_config
    fn_lst_filters = os.path.join(TEST_DATA_DIR, "test_list_filtering_config.json")
    with open(fn_lst_filters, "r") as fid:
        lst_filters = json.load(fid)

    base_function = lambda a, b: a.shape[0]
    result = decorator(lst_filters)(base_function)(M, nrn)
    assert len(result) == len(lst_filters)
    assert (nrn["etype"] == 2).sum() + len(nrn) == result.sum()  # Overlap in etype==2

    for fltr in lst_filters: fltr.pop("name")
    result = decorator(lst_filters)(base_function)(M, nrn)
    assert len(result) == len(lst_filters)
    assert (nrn["etype"] == 2).sum() + len(nrn) == result.sum()  # Overlap in etype==2


def test_analysis_with_control_sample():
    C = ConnectivityMatrix(M, vertex_properties=nrn)
    res = C.analyze(os.path.join(TEST_DATA_DIR, "test_analysis_control_sample.json"))
    sc = res["simplex_counts"]
    assert "sampled_by_mtype" in sc
    assert ("sampled_by_mtype", "in_assembly2") in sc
    assert ("sampled_by_mtype", "in_assembly2", 0) in sc



# SPDX-License-Identifier: Apache-2.0
import numpy


class MorphologyPathDistanceCalculator(object):
    """
    A helper to quickly calculate path distances along the morphology of a neuron. 
    
    Relies on pre-calculating a lot of data and offsets for a single morphology. This is only really
    worth the cost when you want to look up distances between hundreds of points on a single morphology.
    For smaller use cases (or cases where the distances are to be calculated on many different morphologies)
    other code may be faster. 

    NOTE: This class makes the strong assumption that the first section in the morphology is a direct descendant
    of the soma.
    """
    def __init__(self, morphology):
        """
        Args:
          morphology (morphio.Morphology): The morphology you want to calculate path distances for.
        """
        assert 0 in morphology.connectivity[-1], "This class requires the first section to be connected to the soma!"
        self.m = morphology
        self.D, self.R, self.O = self.__initialize_tables__()
    
    @staticmethod
    def segment_offset_table(m):
        """
        Calculate for each segment in each section the distance from the root of the section to the 
        root of the segment. Since different sections have different numbers of segments, all lists
        of offsets are padded with NaN and stacked to a 2d array.

        Args:
          m (morphio.Morphology): The morphogy in question
        
        Returns:
          offset_table (numpy.array): Table of segment offsets. To be indexed as offset_table[sec_id][seg_id].
          Note: sec_id is base-0, i.e. index 0 is NOT the soma. The soma is not listed at all (assumed to be just a point).
        """
        def cumulative_section_length(sec):
            seg_lengths = numpy.sqrt(numpy.sum(numpy.diff(sec.points, axis=0) ** 2, axis=1))
            return [0] + list(numpy.cumsum(seg_lengths))
    
        offset_lists = [cumulative_section_length(_sec)
                        for _sec in m.sections]
        L = numpy.max([len(_x) for _x in offset_lists])
        offset_lists = [_x + [numpy.NaN for _ in range(L - len(_x))]
                       for _x in offset_lists]
        return numpy.vstack(offset_lists)

    def __initialize_tables__(self):
        """
        Calculate the path distances between the roots of pairs of sections, and their relation within the morphology tree.

        Returns:
            root_D_tbl (numpy.array): The table of pairwise path distances between the roots of all sections of the morphology.
            relation_tbl (numpy.array): The table of tree-relations of all pairs of sections. Entries are integers with the following
            meaning:
              -1: non-initialized; this should never happen
              0: Same section; i.e. this should only occur along the main diagonal
              1: tbl[i, j] == 1 means section j is the ancestor of section i
              2: tbl[i, j] == 2 means section j is the descendant of section i
              3: tbl[i, j] == 3 means section i and j are siblings
            O (numpy.array): See segment_offset_table
        """
        m = self.m
        O = self.segment_offset_table(m)
        L = numpy.nanmax(O, axis=1)

        root_D_tbl = -numpy.ones((len(m.sections), len(m.sections)))
        root_D_tbl[numpy.diag_indices_from(root_D_tbl)] = 0.0
        relation_tbl = -numpy.ones((len(m.sections), len(m.sections)), dtype=int) # -1 non-init; 0 same; 1 j ancestor; 2 j descendant; 3 siblings
        relation_tbl[numpy.diag_indices_from(relation_tbl)] = 0

        def recursive_fill(called_sec, parent, with_length):
            called = called_sec.id
            if parent >= 0:
                root_D_tbl[called, parent] = with_length
                root_D_tbl[parent, called] = with_length
                relation_tbl[called, parent] = 1
                relation_tbl[parent, called] = 2

                for j, rows in enumerate(zip(root_D_tbl[parent], relation_tbl[parent])):
                    row_D, row_rel = rows
                    if j == called: continue
                    if row_rel == 1: # generalized node is ancestor of parent node
                        relation_tbl[called, j] = 1
                        root_D_tbl[called, j] = row_D + with_length
                        relation_tbl[j, called] = 2
                        root_D_tbl[j, called] = row_D + with_length
                    elif row_rel == 2: # generalized node is descendant of parent node
                        relation_tbl[called, j] = 3
                        root_D_tbl[called, j] = row_D - with_length
                        relation_tbl[j, called] = 3
                        root_D_tbl[j, called] = row_D - with_length
                    elif row_rel == 3: # generalized node is sibling of parent node
                        relation_tbl[called, j] = 3
                        root_D_tbl[called, j] = row_D + with_length
                        relation_tbl[j, called] = 3
                        root_D_tbl[j, called] = row_D + with_length
            for child in called_sec.children:
                recursive_fill(child, called, L[called])

        for child1 in m.root_sections:
            cid1 = child1.id
            for child2 in m.root_sections:
                cid2 = child2.id
                if cid1 != cid2:
                    root_D_tbl[cid1, cid2] = 0.0
                    relation_tbl[cid1, cid2] = 3
        for child in m.root_sections:
            recursive_fill(child, -1, 0)
        
        return root_D_tbl, relation_tbl, O
    
    def within_section_offsets(self, locs,
                               str_section_id="afferent_section_id",
                               str_segment_id="afferent_segment_id",
                               str_offset="afferent_segment_offset"):
        """
        Calculate the offset of points within a section.

        Args:
          locs (pandas.DataFrame): dataframe where columns specify the section id, segment id and within-segment offset
          of points on the neuron morphology. 
          str_section_id (default="afferent_section_id"): The name of the column holding the section id.
          str_segment_id (default="afferent_segment_id"): The name of the column holding the segment id.
          str_offset (defaults="afferent_segment_offset"): The name of the column holding the within-segment offset.
        
        Note: The default column names are set to what bluepysnap returns when asked for afferent synapses.
        Also, the section id is expected to be base-1 indexed. That is, index 0 is the soma, index 1 is the first "proper" section.
        As path distance along the soma is a tricky concept, the function will return 0.0 for all entries on the soma!

        Returns:
          offsets (numpy.array): The within-section offsets in the same order as the input.
        """
        offsets = self.O[locs[str_section_id].values - 1, locs[str_segment_id]]
        assert not numpy.any(numpy.isnan(offsets))
        offsets = offsets + locs[str_offset].values
        offsets[locs[str_section_id].values == 0] = 0.0
        return offsets
    
    def path_distances(self, locs_from, locs_to=None,
                       str_section_id="afferent_section_id",
                       str_segment_id="afferent_segment_id",
                       str_offset="afferent_segment_offset",
                       same_section_only=False):
        """
        Calculate path distances. Either pairwise, or between two sets of points.
        Args:
          locs_from (pandas.DataFrame): Locations of points on the morphology represented by this object
          to calculate the distances from. A dataframe where columns specify the section id, segment id and within-segment offset
          of points on the neuron morphology.
          locs_to (pandas.DataFrame, default=None): Locations to calculate the distances to. If none, locs_from is used.
          str_section_id (default="afferent_section_id"): The name of the column holding the section id.
          str_segment_id (default="afferent_segment_id"): The name of the column holding the segment id.
          str_offset (defaults="afferent_segment_offset"): The name of the column holding the within-segment offset.
          same_section_only (bool, default=False): If True, set values for all pairs NOT on the same section to NaN.

          Note: The default column names are set to what bluepysnap returns when asked for afferent synapses.
          Also, the section id is expected to be base-1 indexed. That is, index 0 is the soma, index 1 is the first "proper" section.
          This is also compatible with how bluepysnap returns the information.
          Since the concept of path distance along the soma is tricky, this method assumes that the soma is a point.
          Consequently, for all locations on the soma, the values will be as if the location was at the root of a root section.
        
        Returns:
          dist (numpy.array): Array of pairwise path distances.
        """
        if locs_to is None:
            locs_to = locs_from
        
        o_from = self.within_section_offsets(locs_from, str_section_id=str_section_id,
                                             str_segment_id=str_segment_id, str_offset=str_offset)
        o_to = self.within_section_offsets(locs_to, str_section_id=str_section_id,
                                           str_segment_id=str_segment_id, str_offset=str_offset)
        if same_section_only:
            same_section = numpy.eye(len(self.m.sections) + 1, dtype=bool)
            same_section = same_section[numpy.ix_(locs_from[str_section_id], locs_to[str_section_id])]
            dist = numpy.NaN * numpy.ones(same_section.shape, dtype=float)
            i, j = numpy.nonzero(same_section)
            dist[i, j] = numpy.abs(o_from[i] - o_to[j])
            return dist
        
        sec_idx_fr = locs_from[str_section_id].copy() - 1
        sec_idx_to = locs_to[str_section_id].copy() - 1
        # Since we test above that the first section is rooted at the soma, we can use that section in place of the soma.
        # This makes sense, because we assume the soma to be a point (and consequently, within_section_offset) returns
        # 0.0 for all locations on the soma.
        sec_idx_fr[sec_idx_fr < 0] = 0; sec_idx_to[sec_idx_to < 0] = 0

        dist = self.D[numpy.ix_(sec_idx_fr,
                                sec_idx_to)].copy()
        rel = self.R[numpy.ix_(sec_idx_fr,
                                sec_idx_to)]
        
        i, j = numpy.nonzero(rel == 0) # on same section
        dist[i, j] = dist[i, j] + numpy.abs(o_from[i] - o_to[j])

        i, j = numpy.nonzero(rel == 1) # locs_to (j) is ancestor
        dist[i, j] = dist[i, j] + o_from[i] - o_to[j]
        
        i, j = numpy.nonzero(rel == 2) # locs_from (i) is ancestor
        dist[i, j] = dist[i, j] - o_from[i] + o_to[j]
        
        i, j = numpy.nonzero(rel == 3) # siblings
        dist[i, j] = dist[i, j] + o_from[i] + o_to[j]
        return dist
    
    @staticmethod
    def __nnd__(D, _idxx):
        if len(_idxx) <= 1: return numpy.NaN
        return numpy.nanmin(D[numpy.ix_(_idxx, _idxx)], axis=0)
    
    @staticmethod
    def __rnd_sampler__(locs, prop_col, n_rnd):
        import pandas
        if prop_col is None:
            tbl = locs["index"].reset_index(drop=True)
        else:
            tbl = locs.set_index(prop_col)["index"]
        tbl_idx = pandas.Series(numpy.unique(tbl.index))
        def sampler(N):
            for _ in range(n_rnd):
                yield tbl[tbl_idx.sample(N)].values
        return sampler
    
    def nearest_neighbor_distances(self, locs, normalize=False,
                                   normalize_preserve=None,
                                   normalize_n=10,
                                   str_section_id="afferent_section_id",
                                   str_segment_id="afferent_segment_id",
                                   str_offset="afferent_segment_offset",
                                   same_section_only=False):
        """
        Analyzes the nearest neighbor distances between sets of dendritic locations and optionally
        normalizes them against random controls.

        Args:
          locs (pandas.DataFrame): Locations of points on the morphology represented by this object.
          For details, see MorphologyPathDistanceCalculator.path_distances. Additionally, must have
          an index that defines a grouping (see below).
          str_section_id (str, optional): See MorphologyPathDistanceCalculator.path_distances.
          str_segment_id (str, optional): See MorphologyPathDistanceCalculator.path_distances.
          str_offset (str, optional): See MorphologyPathDistanceCalculator.path_distances.
          same_section_only (bool, default: False): See MorphologyPathDistanceCalculator.path_distances.
          normalize (bool, default: False): Whether to normalize the results by converting to a
          z-score with respect to a random control
          normalize_preserve (optional): Name of a column in locs that defines a grouping. 
          When normalizing and this is specified, then locations belonging to the same group will be kept
          together in the random controls. Ignored if normalize=False
          normalize_n (int, default: 10): How many random controls to generate for normalization.
          Ignored if normalize=False.
        
        Explaining this analysis in the context of locations of synapses on dendritic trees and their
        (potential) clustering. 
        The DataFrame locs defines the locations of all afferent synapses onto the morphology in terms of
        section id, segment id and offset. It has an index (can be MultiIndex) that groups them into any
        number of groups by their unique values. This can be the name of the source region, identifier of
        a structural or functional cluster, or simply a bool specifying whether the synapse is of interest
        or not. The analysis then investigates whether the synapses belonging to the same group are 
        clustered, i.e. closer together than expected, or not, and to what degree.

        res = M.nearest_neighbor_distances(locs, normalize=False) simply returns a DataFrame with the
        nearest neighbor distances of synapses in each group. It has one row per unique value of the input
        index, and values are lists with the nearest neighbor distances of all synapses associated with 
        that index.

        res = M.nearest_neighbor_distances(locs, normalize=True) converts the distances into z-scores
        by comparing to a random control. The control is generated by randomly sampling the same number of
        locations from locs. That is, it compares the distances in each group to the distances in random
        groups of the same sizes.

        res = M.nearest_neighor_distances(locs, normalize=True, normalize_preserve="some_property")
        slightly adjusts how the random control is generates. The value of normalize_preserve must be the
        name of a column in locs. Then, in the generation of the controls, locations that have the same
        value in that column will always be sampled together. For example, the column could hold an
        identifier of the presynaptic neuron. Then you can use tis feature to test the clustering of 
        inputs from a neuronal assembly while taking into account that inputs from the same presynaptic
        neuron are already likely to be slightly clustered.

        res = M.nearest_neighbor_distance(..., normalize_n=25) controls how many random controls are
        generated for normalization.

        Examples
        """
        D = self.path_distances(locs, str_section_id=str_section_id,
                                str_segment_id=str_segment_id, str_offset=str_offset,
                                same_section_only=same_section_only)
        numpy.fill_diagonal(D, numpy.NaN)
        index_names = list(locs.index.names)
        assert "index" not in index_names
        locs = locs.reset_index().reset_index()
        data = locs.groupby(index_names)["index"].apply(lambda _s: self.__nnd__(D, _s))
        if not normalize:
            return data
        
        sampler = self.__rnd_sampler__(locs, normalize_preserve, normalize_n)
        if normalize_preserve is None: normalize_preserve = "index"
        counts = locs.groupby(index_names)[normalize_preserve].apply(lambda x: len(x.drop_duplicates()))
        ctrl = counts.apply(lambda c: numpy.hstack([self.__nnd__(D, _s) for _s in sampler(c)]))
        return data.combine(ctrl, lambda _d, _c: (numpy.nanmean(_d) - numpy.nanmean(_c)) / numpy.nanstd(_c))



# SPDX-License-Identifier: Apache-2.0
import numpy
import pandas

from scipy import sparse
from scipy.spatial import KDTree

from .circuit_models.neuron_groups import group_by_grid
from .circuit_models.neuron_groups import load_filter
from .circuit_models.neuron_groups.defaults import GID
from .circuit_models import circuit_connection_matrix, full_connection_matrix


COLS_NAN = ["ss_flat_x", "ss_flat_y"]


def multi_scale_grouping(nrn, radii, properties=["ss_flat_x", "ss_flat_y"]):
    radii = sorted(radii)
    nrn = nrn.reset_index()
    for i, radius in enumerate(radii):
        nrn = group_by_grid(nrn, properties, radius=radius,
                            prefix="level{0}-".format(i + 1)).reset_index()
    #nrn = nrn.rename(columns=dict(zip(properties, ["level0-x", "level0-y"])))
    #nrn["level0-subtarget"] = numpy.arange(len(nrn))

    return nrn


def count_blocks_of_sparse_matrix(M, szs):
    L = len(szs)
    splts = numpy.hstack([0, numpy.cumsum(szs)])
    M = M.tocsc()
    shape = M.shape
    M = sparse.csc_matrix((M.data, M.indices, M.indptr[splts]), shape=(shape[0], L))
    M = M.tocsr()
    M = sparse.csr_matrix((M.data, M.indices, M.indptr[splts]), shape=(L, L))
    return M


class MultiScaleConnectome(object):

    def __init__(self, extent, children, **kwargs):
        self._extent = extent
        self._children = children
        self._props = kwargs
    
    def skip_and_collapse(self, must_be_balanced=True):
        if not self.isleaf:
            if numpy.any([child.isleaf for child in self._children]):
                if must_be_balanced:
                    assert numpy.all([child.isleaf for child in self._children])
                self._children = self.idx
            else:
                grandchildren = []
                for child in self._children:
                    grandchildren.extend(child._children)
                self._children = grandchildren
                [child.skip_and_collapse(must_be_balanced=must_be_balanced)
                for child in self._children]

    @property
    def isleaf(self):
        return isinstance(self._children, numpy.ndarray)
    
    @property
    def idx(self):
        if self.isleaf:
            return self._children
        return numpy.hstack([child.idx for child in self._children])
    
    @property
    def depth(self):
        if self.isleaf:
            return 0
        return numpy.max([x.depth for x in self._children]) + 1
    
    def count(self, at_reach=None):
        if at_reach == 0:
            return 1
        if self.isleaf:
            return len(self._children)
        if at_reach is None:
            return numpy.sum([x.count() for x in self._children])
        return numpy.sum([x.count(at_reach=at_reach - 1) for x in self._children])
    
    def evaluate_at_depth(self, func, depth):
        selfdepth = self.depth
        assert selfdepth >= depth and depth >= 0
        if selfdepth == depth:
            return [func(self)]
        ret = []
        for child in self._children:
            ret.extend(child.evaluate_at_depth(func, depth))
        return ret
    
    def nrn(self):
        if self.isleaf:
            return self._children
        return pandas.concat([x.nrn() for x in self._children])
    
    @staticmethod
    def __nearest_neighbor_interpolation_for_nans__(nrn, cols_use=["x", "y", "z"],
                                                    add_noise=1.0):
        iv = numpy.any(numpy.isnan(nrn[COLS_NAN]), axis=1)
        K = KDTree(nrn[~iv][cols_use])
        D, idx = K.query(nrn[iv][cols_use])
        print("Interpolation of NaN locations: {0} neurons; distance: {1}, {2}, {3} (min/mean/max)".format(len(D), D.min(), D.mean(), D.max()))
        nrn.loc[iv, COLS_NAN] = nrn.loc[~iv, COLS_NAN].iloc[idx].values + numpy.random.rand(len(idx), len(COLS_NAN)) * add_noise - add_noise / 2
        return nrn

    @classmethod
    def from_circuit(cls, circ, load_cfg_or_df, leafsize=100, nan_policy="interpolate"):
        if isinstance(load_cfg_or_df, dict):
            nrn = load_filter(circ, load_cfg_or_df)
        else:
            nrn = load_cfg_or_df
        if nan_policy == "interpolate":
            nrn = cls.__nearest_neighbor_interpolation_for_nans__(nrn)
        elif nan_policy == "drop":
            nrn = nrn.loc[~numpy.any(numpy.isnan(nrn[COLS_NAN]), axis=1)]
        elif numpy.any(numpy.isnan(nrn[COLS_NAN])):
            raise ValueError("Unknown nan_policy: {0}".format(nan_policy))

        nrn = nrn.sort_values("gid")

        data = nrn[["ss_flat_x", "ss_flat_y"]]
        bbox = list(zip(data.min(), data.max()))
        T = KDTree(nrn[["ss_flat_x", "ss_flat_y"]], leafsize=leafsize)
        
        def _recursive(t, bbox, offset=0):
            if isinstance(t, T.innernode):
                bbox_gr = [
                    lim if i != t.split_dim
                    else (t.split, lim[1]) for i, lim in enumerate(bbox)
                ]
                bbox_l = [
                    lim if i != t.split_dim
                    else (lim[0], t.split) for i, lim in enumerate(bbox)
                ]
                children = []
                child, offset_out = _recursive(t.greater, bbox_gr, offset=offset)
                children.append(child)
                child, offset_out = _recursive(t.less, bbox_l, offset=offset_out)
                children.append(child)
                return cls(bbox, children, offset=offset), offset_out
            return cls(bbox, t.idx, offset=offset), offset + len(t.idx)
        ret, _ = _recursive(T.tree, bbox)
        ret._props["neurons"] = nrn
        ret._props["circuit"] = circ
        return ret
    
    def __attach_matrices__(self, M, tgt_range=10):
        n_node = numpy.mean(self.evaluate_at_depth(lambda x: x.count(), 0))
        r_node = int(numpy.floor(numpy.log2(n_node)))

        idx = self.idx
        nrn_df = self._props["neurons"][["ss_flat_x", "ss_flat_y"]]
        M = M[:, idx][idx]  # Now the rows/cols are in the order given by the multi-scale structure

        def assign_neuron_resolution_mat(mat, blck_lens, reinitialize=False):
            blck_off = numpy.hstack([0, numpy.cumsum(blck_lens)]).tolist()
            assert mat.shape[0] == blck_off[-1]
            def out_func(node):
                fr = blck_off[0]; to = blck_off[1]
                blck_off.pop(0)
                node._props["matrix"] = mat[numpy.ix_(range(fr, to), range(fr, to))]
                if reinitialize:
                    node._props["matrix"] = node._props["matrix"].tocoo().tocsr()
            return out_func
        
        def node_locations_at_resolution(depth_resolution):
            def out_func(node):
                res = node.evaluate_at_depth(lambda x: list(map(numpy.mean, x._extent)), depth_resolution)
                return numpy.vstack(res)
            return out_func
        
        def assign_results_to(func, name_str):
            def out_func(node):
                node._props[name_str] = func(node)
            return out_func

        # Neuron level resolution
        print("At resolution level: neurons")
        lengths_at_sampling = self.evaluate_at_depth(lambda x: x.count(at_reach=tgt_range), tgt_range - r_node)
        _ = self.evaluate_at_depth(assign_results_to(lambda x: nrn_df.iloc[x.idx].values,
                                                     "node_locations"), tgt_range - r_node)
        _ = self.evaluate_at_depth(assign_neuron_resolution_mat(M, lengths_at_sampling), tgt_range - r_node)

        for curr_res in range(0, self.depth - tgt_range + 1):
            print("At resolution level: {0}".format(curr_res))
            lengths_at_resolution = self.evaluate_at_depth(lambda x: x.count(at_reach=1), curr_res)
            _ = self.evaluate_at_depth(assign_results_to(node_locations_at_resolution(curr_res),
                                                         "node_locations"), tgt_range + curr_res)
            lengths_at_sampling = self.evaluate_at_depth(lambda x: x.count(at_reach=tgt_range), tgt_range + curr_res)
            M = count_blocks_of_sparse_matrix(M, lengths_at_resolution)
            _ = self.evaluate_at_depth(assign_neuron_resolution_mat(M, lengths_at_sampling, reinitialize=True),
                                                                    tgt_range + curr_res)
        
    def __remove_unattached_nodes__(self):
        if not self.isleaf:
            children = []
            array_flag = False
            for child in self._children:
                r = child.__remove_unattached_nodes__()
                array_flag = array_flag or isinstance(r, numpy.ndarray)
                children.extend(r)
            if array_flag:
                self._children = numpy.array(children)
            else:
                self._children = children

        if "matrix" in self._props:
            return [self]
        return self._children

    def attach_matrices(self, matrix_kwargs={}, tgt_range=10, force_full_matrix=False):
        circ = self._props["circuit"]
        gids = self._props["neurons"][GID]
        if force_full_matrix:
            assert len(gids) == circ.nodes.size

        if not isinstance(matrix_kwargs, list):
            matrix_kwargs = [matrix_kwargs]
        M = sparse.csc_matrix((len(gids), len(gids)), dtype=float)
        for m_kw in matrix_kwargs:
            if force_full_matrix:
                Madd = circuit_connection_matrix(circ, **m_kw)
            else:
                Madd = circuit_connection_matrix(circ, for_gids=gids, **m_kw)
            if isinstance(Madd, dict):
                assert len(Madd) == 1
                Madd = list(Madd.values())[0]
            M = M + Madd
        self.__attach_matrices__(M, tgt_range=tgt_range)
        self.__remove_unattached_nodes__()
    
    def to_h5(self, fn, group="ms_matrix"):
        import h5py
        with h5py.File(fn, "w") as h5_file:
            h5 = h5_file.require_group(group)
            grp_pattern = "node{0}"

            def __recursive__(n, node_index):
                grp = h5.create_group(grp_pattern.format(node_index))
                grp.create_dataset("matrix", data=numpy.array(n._props["matrix"].todense()))
                grp.create_dataset("node_locations", data=n._props["node_locations"])
                grp.attrs["extent_x"] = n._extent[0]
                grp.attrs["extent_y"] = n._extent[1]

                child_ids = []
                if n.isleaf:
                    grp.create_dataset("neuron_indices", data=n._children)
                else:
                    for child in n._children:
                        child_ids.append(node_index + 1)
                        node_index = __recursive__(child, node_index + 1)
                grp.attrs["child_ids"] = child_ids
                
                return node_index
            __recursive__(self, 0)
            root_grp_name = grp_pattern.format(0)
        self._props["neurons"].to_hdf(fn, key=group + "/" + root_grp_name + "/neurons", format="table")



# SPDX-License-Identifier: Apache-2.0
import os
import pandas
import numpy

import bluepysnap as snap

from scipy import sparse
from conntility import connectivity as test_module

from utils import get_snap_test_circuit

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_DATA_DIR = os.path.join(TEST_DIR, "data")

numpy.random.seed(123)
m_dense = numpy.random.rand(10, 10) < 0.1
m_sparse = sparse.coo_matrix(m_dense)

props = pandas.DataFrame({
    "foo": ["bar"] * 5 + ["meh"] * 5,
    "depth": numpy.linspace(-200, 500, 10)
}, index=numpy.random.permutation(10))


def test_from_bluepy():
    load_cfg = {
        "loading": {
            "properties": ["x", "y", "z", "morphology", "layer"]
            },
            "filtering": {"column": "layer", "value": 6}
            }
    with get_snap_test_circuit() as circ_fn:
        CIRC = snap.Circuit(circ_fn)
        M = test_module.ConnectivityMatrix.from_bluepy(CIRC, load_config=load_cfg,
                                                       connectome="default")
        um = numpy.unique(M.morphology)
        ul = numpy.unique(M.layer)
        assert len(um) > 1 and um[0].startswith("morph")
        assert len(ul) == 1 and ul[0] == 6

def test_base_connectivity_matrix():
    M = test_module.ConnectivityMatrix(m_sparse, vertex_properties=props)

    assert M.matrix.nnz == 6
    assert M.subpopulation([4, 5, 6, 7]).matrix.nnz == 2
    assert M.dense_matrix.sum() == 6
    assert "depth" in M.vertex_properties
    assert "meh" in M.foo


def test_subpopulation():
    M = test_module.ConnectivityMatrix(m_sparse, vertex_properties=props)
    idx = numpy.random.choice(len(M), int(0.8*len(M)), replace=False)
    idgids = M.gids[idx]
    assert (M.subpopulation(idgids).array == M.array[numpy.ix_(idx, idx)]).all()
    assert (M.subpopulation(idgids).array == M.subarray(idgids)).all()


def test_connectivity_matrix_node_indexing():
    M = test_module.ConnectivityMatrix(m_sparse, vertex_properties=props)
    assert len(M.index("depth").lt(300)) == 7
    assert len(M.index("depth").le(300)) == 7
    assert len(M.index("depth").gt(300)) == 3
    assert len(M.index("depth").ge(300)) == 3

    assert len(M.index("foo").eq("meh")) == 5
    assert len(M.index("foo").eq("meh")) == 5
    assert M.index("foo").isin(["bar", "not"]).matrix.nnz == 2

    numpy.random.seed()
    seed = numpy.random.randint(0, 100000)
    sel_gids = numpy.random.choice(M.gids[M.depth < M.depth[-1]], 4, replace=False)
    numpy.random.seed(seed)
    gids = M.index("depth").random_numerical_gids(sel_gids, n_bins=2)
    assert M.gids[-1] not in gids
    assert len(gids) == len(sel_gids)
    numpy.random.seed(seed)
    subpop = M.index("depth").random_numerical(sel_gids, n_bins=2)
    assert numpy.all(gids == subpop.gids)

    sel_gids = M.gids[[0, -1]]
    subpop = M.index("foo").random_categorical(sel_gids)
    assert len(subpop) == 2
    assert "bar" in subpop.foo and "meh" in subpop.foo


def test_connectivity_matrix_edge_indexing_and_adding():
    M = test_module.ConnectivityMatrix(m_sparse, vertex_properties=props)
    assert M.filter().lt(2).matrix.nnz == M.matrix.nnz
    assert M.filter().lt(1).matrix.nnz == 0

    numpy.random.seed(123)
    M.add_edge_property("for_testing", numpy.random.rand(len(M._edges)))
    Md = M.default("for_testing")
    assert len(Md.filter().lt(0.5)._edges) + len(Md.filter().ge(0.5)._edges) == len(M._edges)
    assert len(Md.filter().le(0.5)._edges) + len(Md.filter().gt(0.5)._edges) == len(M._edges)
    assert M.default("for_testing").filter().lt(0.5).matrix.nnz == 3

    int_property = numpy.random.randint(1, 6, size=M.array.shape)
    int_property[M.array == 0] = 0
    M.add_edge_property("category_testing", int_property)
    M.add_edge_property("category_testing2", sparse.coo_matrix(int_property))
    assert len(M.filter("category_testing").isin([1, 2, 3, 4, 5])._edges) == len(M._edges)
    fltrd_M = M.filter("category_testing").eq(1)
    assert (fltrd_M.edges["category_testing"] == 1).mean() == 1.0


def test_load_save():
    M = test_module.ConnectivityMatrix(m_sparse, vertex_properties=props)
    M.to_h5("test.h5", group_name="test_group", prefix="test_matrix")
    N = test_module.ConnectivityMatrix.from_h5("test.h5", group_name="test_group",
    prefix="test_matrix")
    assert (M.array == N.array).all()


def test_time_dependent_matrix():
    numpy.random.seed(123)
    row = numpy.random.randint(0, 100, 500)
    col = numpy.random.randint(0, 100, 500)
    rowcol = pandas.DataFrame({"row": row, "col": col}).drop_duplicates()
    L = len(rowcol)
    df = {
        "a": pandas.DataFrame(numpy.random.rand(L, 3), columns=[0.0, 10.0, 20.0]),
        "b": pandas.DataFrame(numpy.random.rand(L, 3), columns=[0.0, 10.0, 20.0])
    }

    M = test_module.TimeDependentMatrix(rowcol["row"].values, rowcol["col"].values,
                                        edge_properties=df)
    assert M.matrix.nnz == len(rowcol)
    assert M.edges.shape[1] == 2
    assert M._time == 0.0
    assert M.filter().lt(0.2).matrix.nnz == 82
    assert M.at_time(10.0).filter().lt(0.2).matrix.nnz == 98
    Mb = M.default("b")
    assert Mb._time == M._time
    assert Mb.at_time(20.0).filter().lt(0.2).matrix.nnz == 96

def test_random_edge_sampling():
    N = 500
    E = int(0.05 * N * (N - 1))

    edge_ids = pandas.DataFrame.from_dict(
        {
            "row": numpy.random.randint(0, N, size=E),
            "col": numpy.random.randint(0, N, size=E)
        }
    ).drop_duplicates()

    vertex_props = pandas.DataFrame(
        numpy.random.randint(0, 5, size=(N, 3)),
        columns=["foo", "bar", "meh"],
        index=pandas.RangeIndex(200, 200 + N)
    )
    edge_properties = numpy.random.rand(len(edge_ids))
    edge_properties[vertex_props["foo"].iloc[edge_ids["row"]] < 2] += 100

    M = test_module.ConnectivityMatrix(edge_ids, vertex_properties=vertex_props,
                                       edge_properties=edge_properties)
    
    subpop = M.index("foo").lt(2)
    assert subpop.edges[subpop._default_edge].mean() > 99
    ctrlpop = M.filter().random_by_vertex_property(subpop, "bar")
    ctrlpop2 = M.filter().random_by_vertex_property_ids(subpop.gids, "bar")
    ctrlpop2 = M.subedges(ctrlpop2)
    assert (ctrlpop.edges.mean() < subpop.edges.mean()).all()  # Strictly not 100% guaranteed, but likely
    assert (ctrlpop2.edges.mean() < subpop.edges.mean()).all()

    A = subpop.edge_associated_vertex_properties("bar").value_counts()
    B = ctrlpop.edge_associated_vertex_properties("bar").value_counts()
    C = ctrlpop2.edge_associated_vertex_properties("bar").value_counts()
    assert (A == B).all() and (A == C).all()

    subpop = M.filter().lt(90)
    eavp = subpop.edge_associated_vertex_properties("foo")
    assert not (eavp["row"] < 2).any()
    manual_fltr = M.edges[M._default_edge] < 90
    ctrlpop = M.filter().random_by_vertex_property(subpop, "meh")
    ctrlpop2 = M.filter().random_by_vertex_property(manual_fltr, "meh")

    A = subpop.edge_associated_vertex_properties("meh").value_counts()
    B = ctrlpop.edge_associated_vertex_properties("meh").value_counts()
    C = ctrlpop2.edge_associated_vertex_properties("meh").value_counts()
    assert (A == B).all() and (A == C).all()



# SPDX-License-Identifier: Apache-2.0
# Defines a number of default file names, column names and other strings
VIRTUAL_FIBERS_FN = "virtual-fibers.csv"

XYZ = ["x", "y", "z"]
UVW = ["u", "v", "w"]
FLAT_COORDINATES = ["flat_x", "flat_y"]
SS_COORDINATES = ["ss_flat_x", "depth", "ss_flat_y"]
GID = "node_ids"
FIBER_GID = "sgid"
PROJECTION = "projection"
DEFAULT_EDGES = "default"
DEFAULT_PIXEL_SZ = 34.0



# SPDX-License-Identifier: Apache-2.0
import numpy

from voxcell import VoxcellError

FIX_TRANSITION = 1500  # In the SSCx Bio_M the projection fiber positions are defined at the start of the rays,
# which are outside of Sirio's flatmap. They have to be shifted with 1500 (as of 11.2021) to get them into the flatmap.


def colored_points_to_image(flat_coords, cols, extent=None):
    flat_coords = flat_coords.astype(int)
    if extent is None:
        extent = flat_coords.max(axis=0) + 1
    img = numpy.NaN * numpy.ones(tuple(extent) + (3, ))
    for xy, col in zip(flat_coords, cols):
        img[xy[0], xy[1], :] = col
    return img


def _flatmap_extent(fm, subsample=None):
    mx = fm.raw.max(axis=(0, 1, 2))
    if subsample is not None:
        mx = numpy.floor(mx / subsample).astype(int)
    return mx + 1


def _flat_coordinates_of_regions(names_region, fm, hier, ann, make_unique=True, subsample=None):
    reg_ids = set()
    for region in names_region:
        reg_ids.update(hier.find(region, "acronym", with_descendants=True))
    lst_ids = list(reg_ids)
    view3d = numpy.in1d(ann.raw.flat, lst_ids).reshape(ann.raw.shape)
    view2d = fm.raw[view3d]
    view2d = view2d[numpy.all(view2d >= 0, axis=1)]

    if subsample is not None:
        view2d = numpy.floor(view2d / subsample).astype(int)

    if not make_unique:
        return view2d
    mul = view2d[:, 0].max() + 1
    view_comb = view2d[:, 0] + mul * view2d[:, 1]
    unique_comb = numpy.unique(view_comb)
    return numpy.vstack([
        numpy.mod(unique_comb, mul),
        numpy.floor(unique_comb / mul)
    ]).transpose()


def flat_coordinates_of_regions(names_regions, fm, *args, make_unique=False, subsample=None):
    if len(args) == 2:
        return _flat_coordinates_of_regions(names_regions, fm, *args, make_unique=make_unique, subsample=subsample)
    elif len(args) == 1:
        circ = args[0]
        from ..circuit_models.neuron_groups.sonata_extensions import load_atlas_data, load_atlas_hierarchy
        ann = load_atlas_data(circ, "brain_regions")
        hier = load_atlas_hierarchy(circ)
        return _flat_coordinates_of_regions(names_regions, fm, hier, ann, make_unique=make_unique, subsample=subsample)
    else:
        raise ValueError()


def flat_region_image(lst_regions, fm, *args, extent=None, subsample=None):
    if extent is None:
        extent = _flatmap_extent(fm, subsample=subsample)
    xbins = numpy.arange(extent[0] + 1)
    ybins = numpy.arange(extent[1] + 1)
    counters = []
    for reg in lst_regions:
        if not isinstance(reg, list):
            reg = [reg]
        A = flat_coordinates_of_regions(reg, fm, *args, make_unique=False, subsample=subsample)
        H = numpy.histogram2d(A[:, 0], A[:, 1], bins=(xbins, ybins))[0]
        counters.append(H)

    region_lookup = numpy.argmax(numpy.dstack(counters), axis=2)
    region_lookup[numpy.sum(numpy.dstack(counters), axis=2) == 0] = -1
    return region_lookup


def apply_flatmap_with_translation(xyz, uvw, fm, max_translation=2000):
    """
    Uses Voxcell to look up locations of `xyz` in the flatmap. If locations are outside the valid region of
    the flatmap (which is usually the case for projections as those start at the bottom of L6)
    and `uvw` is provided (i.e. not `None`), then there are 2 possible options:
    1) The invalid locations are translated along the directions given by `uvw` by a hard-coded fix amount, or
    2) The invalid locations are gradually translated along the directions given by `uvw` until they hit
    the valid volume. `max_translation` defines the maximum amplitude of that translation.
    Locations that never hit the valid volume will return a flat location of (-1, -1).
    :param xyz: numpy.array, N x 3: coordinates in 3d space
    :param uvw: numpy.array, N x 3: directions in 3d space. Optional, can be None.
    :param fm: VoxelData: flatmap
    :param max_translation: float.
    :return: Flat locations of the xyz coordinates in the flatmap.
    """
    solution = fm.lookup(xyz)
    if uvw is not None and not numpy.all(solution > 0):
        # 1)
        solution = fm.lookup(xyz + FIX_TRANSITION * uvw)
        if numpy.all(solution > 0):
            return solution
        else:
            # 2)
            fac = 0
            step = fm.voxel_dimensions[0] / 4
            tl_factors = numpy.zeros((len(uvw), 1))
            solution = fm.lookup(xyz)
            while numpy.any(solution < 0) and fac < max_translation:
                try:
                    fac += step
                    to_update = numpy.any(solution < 0, axis=1)
                    tl_factors[to_update, 0] = fac
                    solution[to_update, :] = fm.lookup(xyz[to_update, :] + tl_factors[to_update, :] * uvw[to_update, :])
                except VoxcellError:
                    break
    return solution



# SPDX-License-Identifier: Apache-2.0
from . import analysis, circuit_models, flatmapping, io, randomization, subcellular
from .connectivity import ConnectivityMatrix, TimeDependentMatrix, ConnectivityGroup
from .connectivity import StructurallyPlasticMatrix



# SPDX-License-Identifier: Apache-2.0
# Functionality for loading information about neurons in a Circuit
# Wrapper and additional functionality for Circuit.cells.get
import numpy
import pandas
import os

from .extra_properties import add_extra_properties
from .defaults import GID, FIBER_GID, PROJECTION, DEFAULT_EDGES
from ..sonata_helpers import source_connectomes_for, target_connectomes_for


def load_neurons(circ, properties, base_target=None, node_population=None, **kwargs):
    """
    Loads information about the neurons in a Circuit. Provides access to all properties
    that Circuit.cells.get loads, plus flat mapped locations with the following names:
    flat_x: Flat mapped x coordinate (pixel resolution)
    flat_y: Flat mapped y coordinate (pixel resolution)
    ss_flat_x: Supersampled flat mapped x coordinate (approximately in um)
    ss_flat_y: Supersampled flat mapped y coordinate (approximately in um)
    depth: Approximate cortical depth of a neuron in um

    Input:
    circ (bluepysnap.Circuit): The Circuit to load neurons from
    properties (list-like): List with names of properties to load. Can contain anything
    that Circuit.cells.get loads, plus any of the additional properties listed above.
    base_target (str, optional): The name of the neuron target to load. If not provided
    loads all neurons.
    node_population (str, optional): The name of the node population to load from. Currently,
    in multi-population circuits, it is likely required to specify the population.
    """
    props_to_load = numpy.intersect1d(properties, list(circ.nodes.property_names))
    extra_props = numpy.setdiff1d(properties, list(circ.nodes.property_names))
    node = circ.nodes
    if node_population is not None:
        neurons = node[node_population].get(group=base_target, properties=props_to_load)
        neurons.index.name = GID  # Until my pull request is merged.
    else:
        try:
            neurons = node.get(group=base_target, properties=props_to_load)
        except Exception as err:
            if str(err).startswith("Same property"):
                raise ValueError("In multi-population circuits, must use node_population= kwarg")
            raise
    neurons = neurons.reset_index()

    neurons = add_extra_properties(neurons, circ, extra_props, **kwargs)
    return neurons


def load_source_locations(circ, properties, projection_name, 
                          base_target=None, **kwargs):
    """
    Loads anatomical information about source nodes of a connectome.
    Properties that can be loaded: See load_neurons.

    Args:
        circ (bluepysnap.Circuit): The Circuit to load fibers from
        properties (list-like): List with names of properties to load. See load_neurons.
        projection_name (str): The name of the connectome to load the source population
        of. Must be in circ.edges.
        base_target (optional): Name of a node_set. If provided, loads the intersection
        of the source population with the node set.
    """
    if projection_name not in circ.edges:
        raise ValueError("No connectome {0} found!".format(projection_name))
    nodepop = circ.edges[projection_name].source.name
    return load_neurons(circ, properties, base_target=base_target,
                        node_population=nodepop, **kwargs)

def load_target_locations(circ, properties, projection_name, 
                          base_target=None, **kwargs):
    """
    Loads anatomical information about target nodes of a connectome.
    Properties that can be loaded: See load_neurons.

    Args:
        circ (bluepysnap.Circuit): The Circuit to load fibers from
        properties (list-like): List with names of properties to load. See load_neurons.
        projection_name (str): The name of the connectome to load the target population
        of. Must be in circ.edges.
        base_target (optional): Name of a node_set. If provided, loads the intersection
        of the target population with the node set.
    """
    if projection_name not in circ.edges:
        raise ValueError("No connectome {0} found!".format(projection_name))
    nodepop = circ.edges[projection_name].target.name
    return load_neurons(circ, properties, base_target=base_target,
                        node_population=nodepop, **kwargs)


def load_all_source_locations(circ, properties, node_population,
                              base_target=None, **kwargs):
    """
    Loads anatomical information about nodes innervating a node_population.
    Properties that can be loaded: See load_neurons.

    Args:
        circ (bluepysnap.Circuit): The Circuit to load fibers from
        properties (list-like): List with names of properties to load. See load_neurons.
        node_population (str): The name of a node population. All node populations innervating
        this population through any EdgePopulation will be loaded.
        base_target (optional): Name of a node_set. If provided, loads the intersection
        of the above with the node set.
    """
    connectomes = source_connectomes_for(circ, node_population)
    return pandas.concat([
        load_source_locations(circ, properties, conn, base_target=base_target, **kwargs)
        for conn in connectomes
    ], axis=0)


def load_all_target_locations(circ, properties, node_population,
                              base_target=None, **kwargs):
    """
    Loads anatomical information about nodes innervated by a node_population.
    Properties that can be loaded: See load_neurons.

    Args:
        circ (bluepysnap.Circuit): The Circuit to load fibers from
        properties (list-like): List with names of properties to load. See load_neurons.
        node_population (str): The name of a node population. All node populations innervated by
        this population through any EdgePopulation will be loaded.
        base_target (optional): Name of a node_set. If provided, loads the intersection
        of the above with the node set.
    """
    connectomes = target_connectomes_for(circ, node_population)
    return pandas.concat([
        load_source_locations(circ, properties, conn, base_target=base_target, **kwargs)
        for conn in connectomes
    ], axis=0)



# SPDX-License-Identifier: Apache-2.0
from .make_groups import group_by_properties, group_by_grid, group_by_binned_properties
from .loader import load_neurons, load_all_source_locations, load_all_target_locations
from .loader import load_source_locations, load_target_locations
from .defaults import SS_COORDINATES, FLAT_COORDINATES
from .make_groups import flip, count_overlap
from .grouping_config import load_group_filter, load_filter, load_with_config, group_with_config, filter_with_config



# SPDX-License-Identifier: Apache-2.0
"""
Contributed by Vishal Sood
Last changed: 2021/11/29
"""
import importlib
from collections import OrderedDict
from pathlib import Path

import pandas as pd

def import_module(from_path, with_method=None):
    """..."""
    path = Path(from_path)

    assert path.exists

    assert path.suffix == ".py", f"Not a python file {path}!"

    spec = importlib.util.spec_from_file_location(path.stem, path)

    module = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(module)

    if with_method:
        if not hasattr(module, with_method):
            raise TypeError(f"No method to {with_method}")
        return (module, getattr(module, with_method))
    return module


def import_module_with_name(n):
    """Must be in the environment."""
    assert isinstance(n, str)
    return importlib.import_module(n)


def load_module_from_path(p):
    """Load a module from a path.
    """
    path = Path(p)

    if not path.exists:
        raise FileNotFoundError(p.as_posix())

    if  path.suffix != ".py":
        raise ValueError(f"Not a python file {path}!")

    spec = importlib.util.spec_from_file_location(path.stem, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module) #

    return module


def get_module(from_object, with_function=None):
    """Get module from an object.
    Read the code to see what object can be resolved to a module.
    If `with_method`, look for the method in the module.
    """
    def iterate(functions):
        """..."""
        if isinstance(functions, str):
            return [functions]
        try:
            items = iter(functions)
        except TypeError:
            items = [functions]

        return items

    def check(module, has_function=None):
        """..."""
        if not has_function:
            return module

        def get_method(function):
            """..."""
            try:
                method = getattr(module, function)
            except AttributeError:
                raise TypeError(f" {module} is missing required method {function}.")
            return method

        if isinstance(has_function, str):
            methods = get_method(has_function)

        methods = {f: get_method(f) for f in iterate(has_function)}

        return (module, methods)

    try:
        module = import_module_with_name(str(from_object))
    except ModuleNotFoundError:
        module = load_module_from_path(p=from_object)
        if not module:
            raise ModuleNotFoundError(f"that was specified by {from_object}")

    return check(module)


def collect_plugins_of_type(T, in_config):
    """..."""
    return {T(name, description) for name, description in items()}



# SPDX-License-Identifier: Apache-2.0
"""
Contributed by Vishal Sood
Modified by Michael W. Reimann
Last changed: 2021/12/15
"""

import os

from ..plugins import import_module

from ..io.logging import get_logger
from ..circuit_models.neuron_groups.grouping_config import _resolve_includes

LOG = get_logger("Topology Pipeline Analysis", "INFO")


def widen_by_index(level, dataframe):
    """Widen a dataframe by an index level."""
    import pandas as pd
    groups = dataframe.groupby(level)
    return pd.concat([g.droplevel(level) for _, g in groups], axis=1,
                     keys=[i for i, _ in groups], names=[level])


def get_analyses(in_config):
    """..."""
    if isinstance(in_config, str) or isinstance(in_config, os.PathLike):
        import json
        _root = os.path.split(os.path.abspath(in_config))[0]
        with open(in_config, "r") as fid:
            in_config = json.load(fid)
    else:
        _root = None
    analyses = in_config["analyses"]
    return collect_plugins_of_type(SingleMethodAnalysisFromSource,
                                   in_config=analyses, resolve_at=_root)


def collect_plugins_of_type(T, in_config, **kwargs):
    """..."""
    return {T(name, description, **kwargs) for name, description in in_config.items()}


class SingleMethodAnalysisFromSource:
    """Algorithms defined as such in the config:

    "analyze-connectivity": {
      "analyses": {
        "simplex_counts": {
          "source": "/gpfs/bbp.cscs.ch/project/proj83/home/sood/analyses/manuscript/topological-analysis-subvolumes/topologists_connectome_analysis/analysis/simplex_counts.py",
          "args": [],
          "kwargs": {},
          "method": "simplex-counts",
          "output": "scalar"
        }
      }
    }

    """
    @staticmethod
    def read_method(description):
        """..."""
        return description.get("method", "shuffle")

    @staticmethod
    def read_source(description):
        """..."""
        return description["source"]

    @staticmethod
    def read_args(description):
        """..."""
        return description.get("args", [])

    @staticmethod
    def read_kwargs(description):
        """..."""
        return description.get("kwargs", {})

    @staticmethod
    def read_output_type(description):
        """..."""
        return description["output"]

    @staticmethod
    def read_collection(description):
        """..."""
        policy = description.get("collect", None)
        if not policy:
            return None
        return policy
    
    @staticmethod
    def read_decorators(description):
        decorators = description.get("decorators", [])
        if not isinstance(decorators, list): return [decorators]
        return decorators

    def __init__(self, name, description, resolve_at=None):
        """..."""
        self._name = name
        self._root = resolve_at
        self._description = _resolve_includes(description, resolve_at=self._root)
        self._source = self.read_source(description)
        self._args = self.read_args(description)
        self._kwargs = self.read_kwargs(description)
        self._method = self.read_method(description)
        self._decoration = self.read_decorators(description)
        self._output_type = self.read_output_type(description)
        self._analysis = self.load(description)
        self._collection_policy = self.read_collection(description)

    @property
    def name(self):
        """..."""
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def quantity(self):
        """To name the column in a dataframe, or an item in a series."""
        return self._description.get("quantity",
                                     self._description.get("method",
                                                           self._analysis.__name__))

    def load(self, description):
        """..."""
        source = self.read_source(description)
        method = self.read_method(description)

        try:
           run = getattr(source, method)
        except AttributeError:
            pass
        else:
            self._module = source
            return run

        if callable(source):
            #TODO: inspect source
            return source
        
        if not os.path.isfile(source):
            if self._root is not None and not os.path.isabs(source):
                source = os.path.join(self._root, source)

        module, method = import_module(from_path=source, with_method=method)
        self._module = module
        return method
    
    def decorate(self, analysis, lst_decorators, **kwargs):
        from . import analysis_decorators
        for decorator in lst_decorators:
            dec_func = getattr(analysis_decorators, decorator["name"])
            args = decorator.get("args", []).copy()

            for i in range(len(args)):  # To be used to insert dynamic arguments to decorators
                if isinstance(args[i], str):
                    if args[i] in kwargs:
                        args.insert(i, kwargs[args.pop(i)])
            
            if "analysis_arg" in decorator:
                args = [
                    SingleMethodAnalysisFromSource(k, v, resolve_at=self._root)
                    for k, v in decorator["analysis_arg"].items()
                ] + args
            analysis = dec_func(*args, **decorator.get("kwargs", {}))(analysis)
        return analysis

    def apply(self, adjacency, node_properties=None, log_info=None):
        """..."""
        from .. import ConnectivityMatrix
        if log_info:
            LOG.info("APPLY %s", log_info)
        decoration_kwargs = {}
        if isinstance(adjacency, ConnectivityMatrix):
            decoration_kwargs["ConnectivityMatrix"] = adjacency
            node_properties = adjacency.vertices
            adjacency = adjacency.matrix.tocsc()
        elif hasattr(adjacency, "matrix"):
            adjacency = adjacency.matrix

        if node_properties is not None:
            assert node_properties.shape[0] == adjacency.shape[0]

        result = self.decorate(self._analysis, self._decoration, **decoration_kwargs)(
            adjacency, node_properties, *self._args, **self._kwargs
        )

        if log_info:
            LOG.info("Done %s", log_info)

        return result

    @staticmethod
    def collect(data):
        """collect data...
        TODO: We could have the scientist provide a collect method.
        """
        return data



# SPDX-License-Identifier: Apache-2.0
from .flatmap_utility import colored_points_to_image, flat_region_image, flat_coordinates_of_regions, apply_flatmap_with_translation
from .supersampling import supersample_flatmap, supersampled_locations
from .supersampling import estimate_flatmap_pixel_size, per_pixel_coordinate_transformation, pandas_flat_coordinate_frame


# SPDX-License-Identifier: Apache-2.0
import numpy

from tqdm import tqdm
from scipy.sparse import issparse


def similarity_matrix(C):
    """Turns a L x N connectivity matrix into a L x L similarity matrix. I.e. evaluates how similar the
    individual _rows_ of the input matrix are.
    Additionally returns a normalized version of the similarity that is relevant for the
    diffusion embedding process."""
    chunk_size = 100000
    C = C.astype(float)

    if issparse(C):
        C = C.tocsr()
        normalize_vals = numpy.array(numpy.sqrt(C.multiply(C).sum(axis=1))).flatten()
        for i, v in enumerate(normalize_vals): # Normalize connectivity
            C.data[C.indptr[i]:C.indptr[i+1]] = C.data[C.indptr[i]:C.indptr[i+1]] / v
        C = (C * C.transpose()).tocoo() # Similarity
        # TODO: After this, C might no longer really have a sparse structure. Might be better to 
        # cast .todense() and continue the non-sparse case from here.
        nrm = numpy.array(C.sum(axis=1)).flatten()
        chunks = numpy.arange(0, C.nnz + chunk_size, chunk_size)
        for ch_fr, ch_to in tqdm(zip(chunks[:-1], chunks[1:]), desc="Normalizing...", total=len(chunks) - 1):
            ratios = nrm[C.row[ch_fr:ch_to]] / nrm[C.col[ch_fr:ch_to]]
            C.data[ch_fr:ch_to] = (C.data[ch_fr:ch_to] / nrm[C.row[ch_fr:ch_to]]) / ratios

        if C.nnz > (0.2 * C.shape[0] * C.shape[1]):
            C = numpy.array(C.todense())
        else:
            C = C.tocsc()
    else:
        normalize_vals = numpy.linalg.norm(C, axis=1, keepdims=True)
        normalize_vals[normalize_vals == 0] = 1E-9 # In case it's zero
        C = C / normalize_vals # Normalize connectivity
        C = numpy.dot(C, C.transpose()) # Similarity
        cs_sum = C.sum(axis=1, keepdims=True)  # L x 1, zero for zero connected voxels
        cs_sum[cs_sum == 0] = 1E-9  # In case it's zero
        cs_sum_ratios = numpy.hstack([cs_sum / _x for _x in cs_sum[:, 0]])
        C = (C / cs_sum) * numpy.sqrt(cs_sum_ratios) # Normalize similarity
    return C


def embed_pathway(C, diffusion_time=1, n_components=3):
    """
    """
    try:
        from mapalign.embed import compute_diffusion_map
    except ImportError:
        raise RuntimeError("""This optional feature requires installation of the mapalign package.
        Obtain it from https://github.com/satra/mapalign""")
    print("Calculating similarity...")
    S_norm = similarity_matrix(C)
    print("...done! Performing diffusion mapping...")
    #  Treat voxels with zero connectivity, which would be considered disconnected from the rest of the graph
    vxl_is_valid = numpy.array(S_norm.sum(axis=1) != 0).flatten()
    embed_coords = numpy.NaN * numpy.ones((S_norm.shape[0], n_components), dtype=float)
    embed_coords[vxl_is_valid, :], embed_res = compute_diffusion_map(S_norm[numpy.ix_(vxl_is_valid, vxl_is_valid)],
                                                                     return_result=True,
                                                                     diffusion_time=diffusion_time,
                                                                     n_components=n_components)
    print("...done!")
    return S_norm, embed_coords, embed_res




# SPDX-License-Identifier: Apache-2.0
# Functionality for tiling locations in a plane
import logging

import pandas as pd
import numpy as np
# from matplotlib import pyplot as plt
# from matplotlib.patches import Rectangle


LOG = logging.getLogger("Flatmap Utility")


def P(x, y):
    """..."""
    return np.array([x, y])


def Array2D(x, y):
    """..."""
    return np.array([x, y])


def Frame2D(**kwargs):
    """..."""
    index = kwargs.pop("index", None)
    dtype = kwargs.pop("dtype", float)
    return pd.DataFrame(kwargs, index=index, dtype=dtype)


def to_cartesian_dataframe(positions):
    """..."""
    return Frame2D(x=positions.rho * np.cos(positions.phi),
                   y=positions.rho * np.sin(positions.phi))


def to_cartesian_position(rho, phi):
    """..."""
    return np.array([rho * np.cos(phi), rho * np.sin(phi)])


def convert_cartesian(arg0, arg1=None):
    """..."""
    if isinstance(arg0, pd.DataFrame):
        assert arg1 is None
        return to_cartesian_dataframe(arg0)

    try:
        rho, phi = arg0
    except TypeError:
        assert arg1 is not None
        return convert_cartesian((arg0, arg1))

    assert arg1 is None

    return to_cartesian_position(rho, phi)


def to_polar_dataframe(positions):
    """..."""
    return Frame2D(rho=np.linalg.norm(positions, axis=1),
                   phi=np.arctan2(positions.y, positions.x))


def to_polar_position(x, y):
    """..."""
    return np.array([np.sqrt(x**2 + y**2),  np.arctan2(y, x)])


def convert_polar(arg0, arg1=None):
    """..."""
    if isinstance(arg0, pd.DataFrame):
        assert arg1 is None
        return to_polar_dataframe(arg0)

    try:
        x, y = arg0
    except TypeError:
        assert arg1 is not None
        return convert_polar((arg0, arg1))

    assert arg1 is None

    return to_polar_position(x, y)


class Line:
    """Help to draw lines..."""
    def __init__(self, through, at_angle, graphic=None):
        """..."""
        self._origin = through
        self._angle = - (2 * np.pi - at_angle) if at_angle >= np.pi else at_angle
        self._slope = np.tan(at_angle)
        self._graphic = graphic

    def y(self, x):
        """..."""
        x0, y0 = self._origin
        return y0 + self._slope * (x - x0)

    def x(self, y):
        """..."""
        x0, y0 = self._origin
        return x0 + (y - y0) / self._slope

    @property
    def angle(self):
        """..."""
        return self._angle

    def plot(self, x0=None, x1=None, y0=None, y1=None,
             fmt=None, **kwargs):
        """..."""
        if self._graphic is None:
            raise TypeError("This Line was defined without a graphic.")

        def resolve(x, y):
            """..."""
            assert not (x is None and y is None),\
                "At least one of x or y should have been non-null."
            if x is not None:
                assert y is None
                y = self.y(x)
            elif y is not None:
                assert x is None
                x = self.x(y)

            return (x, y)

        x0, y0 = resolve(x0, y0)
        x1, y1 = resolve(x1, y1)

        print("plot line ", x0, y0, "to", x1, y1)
        _, axes = self._graphic
        axes.plot([x0, x1], [y0, y1], fmt, **kwargs)

        return self._graphic

    @classmethod
    def connecting(self, point, to_point, in_graphic=None):
        """..."""
        origin  = (point + to_point) / 2.
        dx, dy = to_point - point
        angle = np.arctan2(dy, dx)
        return Line(origin, angle, in_graphic)


def plot_segment(graphic, p0, p1, fmt=None, **kwargs):
    """..."""
    _, axes = graphic
    axes.plot([p0[0], p1[0]], [p0[1], p1[1]], fmt, **kwargs)
    return graphic


class TriTille:
    """A traingular tesselation.
    """
    def __init__(self, side, origin=None, angle=None):
        """
        side : length of the triangle side
        origin : (x, y) location of the origin of the tiling
        angle : angle of the tiling's x-axis w.r.t to the X-axis.
        """
        self._side = side
        self._origin = np.array([0., 0.]) if origin is None else origin
        self._angle  = 0. if angle is None else angle

        self._ratio = np.array([np.sqrt(3.), 1.])

    def translate(self, positions):
        """Translate positions with respect to the origin."""
        return positions - self._origin

    def untranslate(self, positions):
        """..."""
        return positions + self._origin

    def rotate(self, positions):
        """Rotate positions to align with this TriTille's x-axis.
        `positions` are expected to be relative to this origin.
        """
        polar = convert_polar(positions)
        if isinstance(positions, pd.DataFrame):
            return convert_cartesian(polar.assign(phi=(polar.phi - self._angle)))

        rho, phi = polar
        if np.isclose(rho, 0.):
            return np.array([0., 0.])
        return convert_cartesian(np.array([rho, phi - self._angle]))

    def unrotate(self, positions):
        """..."""
        polar = convert_polar(positions)

        if isinstance(positions, pd.DataFrame):
            return convert_cartesian(polar.assign(phi=(polar.phi + self._angle)))

        rho, phi = polar
        if np.isclose(rho, 0.):
            return np.array([0., 0.])
        return convert_cartesian(np.array([rho, phi + self._angle]))

    def relative(self, positions):
        """Position values relative to the origin and x-axis of the tiling.

        positions: (x, y) values
        """
        return self.rotate(self.translate(positions))

    def transform(self, xys):
        """Transform (x, y) positions to (u, v) axii.
        """
        relpos = self.relative(xys) / self._ratio

        u = relpos.x - relpos.y
        v = relpos.x + relpos.y
        return Frame2D(u=u, v=v)#/ self._ratio

    def reverse_transform(self, uvs):
        """Transform (u, v) positions to (x, y).
        """
        relpos = self._ratio * Frame2D(x=(uvs.u + uvs.v) / 2., y=-(uvs.u - uvs.v) / 2.)

        return self.untranslate(self.unrotate(relpos))

    def display(self, gridsize, origin=None, graphic=None,
                return_methods=False, hexgrid=False, **kwargs):
        """Display this triangular lattice.
        """
        gridsize = gridsize or (120, 120)
        try:
            width, height = gridsize
        except TypeError:
            width = height = gridsize

        window = np.array([width, height])

        ori = origin if origin is not None else self._origin

        if graphic is None:
            raise ValueError("Need to specify figure and axes for plotting!")
        else:
            figure, axes = graphic
        graphic = (figure, axes)

        def draw_boundary(color="black", linestyle="-", linewidth=4, label="grid",
                          joinstyle="bevel", padding=None):
            """..."""

            from matplotlib.patches import Rectangle  # TODO: Let's remove this matplotlib dependence
            x0, y0 = ori
            boundary = Rectangle(xy=ori, height=height, width=width,
                                 fill=False, color=color,
                                 linestyle=linestyle, linewidth=linewidth,
                                 joinstyle=joinstyle,
                                 label=label)
            axes.add_patch(boundary)

            try:
                xpad, ypad = padding or self._side
            except TypeError:
                xpad = ypad = padding or self._side

            axes.set_xlim([x0 - xpad, x0 + width + xpad])
            axes.set_ylim([y0 - ypad, y0 + height + ypad])
            return graphic

        def draw_line(through, at_angle, fmt):
            """but only the portion that can be seen through the grid window.
            """
            line = Line(through, at_angle)
            x0, y0 = ori; x1, y1 = ori + window

            line_x0 = x0; line_y0 = line.y(x0)
            if line_y0 < y0:
                if line._slope <= 0:
                    return None
                line_x0 = line.x(y0)
                if line_x0 >= x1:
                    return None
                line_y0 = y0

            if line_y0 > y1:
                if line._slope >= 0.:
                    return None
                line_x0 = line.x(y1)
                if line_x0 >= x1:
                    return None
                line_y0 = y1

            line_x1 = x1; line_y1 = line.y(x1)
            if line_y1 < y0:
                if line._slope >=0:
                    return None
                line_x1 = line.x(y0)
                if line_x1 < x0:
                    return None
                line_y1 = y0

            if line_y1 >= y1:
                if line._slope <= 0:
                    return None
                line_x1 = line.x(y1)
                if line_x1 < x0:
                    return None
                line_y1 = y1

            #print("through", through, "at angle ", at_angle)
            #print("\tsegment", (line_x0, line_y0), (line_x1, line_y1))

            axes.plot([line_x0, line_x1], [line_y0, line_y1], fmt)
            return graphic

        def draw_relxaxis(j, fmt=None):
            """x-axis in this TriTille's reference frame."""
            if hexgrid:
                return None

            in_tritille_coords = P(0, j * self._side)
            in_grid_coords = self.untranslate(self.unrotate(in_tritille_coords))
            return draw_line(in_grid_coords, self._angle, fmt="k--")

        def draw_relyaxis(i, fmt=None):
            """y-axis in this TriTille's reference frame."""
            in_tritille_coords = P(i * self._side, 0.) * self._ratio / 2.
            in_grid_coords = self.untranslate(self.unrotate(in_tritille_coords))
            return draw_line(in_grid_coords, self._angle + np.pi/2, fmt or "k-" if hexgrid else "k--")

        def draw_vaxis(i, fmt="k-"):
            """..."""
            in_tritille_coords = P(0, i * self._side)
            in_grid_coords = self.untranslate(self.unrotate(in_tritille_coords))
            return draw_line(through=in_grid_coords, at_angle=(self._angle + np.pi / 6.), fmt=fmt)

        def draw_uaxis(j, fmt="k-"):
            """..."""
            in_tritille_coords = P(j * self._side, 0.) * self._ratio
            in_grid_coords = self.untranslate(self.unrotate(in_tritille_coords))
            return draw_line(through=in_grid_coords, at_angle=(self._angle - np.pi / 6.), fmt=fmt)

        if return_methods:
            return {"draw_boundary": draw_boundary, "draw_line": draw_line,
                    "draw_relxaxis": draw_relxaxis, "draw_relyaxis": draw_relyaxis,
                    "draw_uaxis": draw_uaxis, "draw_vaxis": draw_vaxis}

        draw_boundary(**kwargs.get("draw_boundary", {}))

        hmax = height + self._origin[1] + 1
        imax = int(hmax / (self._ratio[0] * self._side / 2.))
        vi_outside_grid = [i for i in range(-imax, imax)
                           if not draw_vaxis(i, **kwargs.get("draw_vaxis", {}))]
        LOG.info("Lines along the v-axis that didn't fit the window: %s / %s",
                    len(vi_outside_grid), (2 * imax))
        LOG.info("\t: %s", vi_outside_grid)

        ryi_outside_grid = [i for i in range(-imax, imax)
                            if not draw_relyaxis(i, **kwargs.get("draw_relyaxis", {}))]
        LOG.info("Lines along the TriTille's y-axis that didn't fit the window: %s / %s",
                    len(ryi_outside_grid), (2 * imax))
        LOG.info("\t: %s", ryi_outside_grid)

        wmax = width + self._origin[0] + 1
        jmax = int(wmax / self._side)
        j_outside_grid = [j for j in range(-jmax, jmax)
                          if not draw_uaxis(j, **kwargs.get("draw_uaxis", {}))]

        LOG.info("Lines along the u-axis that didn't fit the window: %s / %s",
                    len(j_outside_grid), 2 * jmax)
        LOG.info("\t: %s", j_outside_grid)

        rxj_outside_grid = [j for j in range(-jmax, jmax)
                            if not draw_relxaxis(j, **kwargs.get("draw_relxaxis", {}))]
        LOG.info("Lines along the TriTille's x-axis that didn't fit the window: %s / %s",
                    len(rxj_outside_grid), (2 * jmax))
        LOG.info("\t: %s", rxj_outside_grid)
        return graphic

    def bin_rhombically(self, xys, **kwargs):
        """..."""
        uvs = self.transform(xys)
        scaled_u = np.array(np.floor(uvs.u.values / self._side), dtype=int)
        scaled_v = np.array(np.floor(uvs.v.values / self._side), dtype=int)
        return Frame2D(i=scaled_u, j=scaled_v, dtype=int, index=xys.index)

    def bin_trinagularly(self, xys, **kwargs):
        """..."""
        ijs = self.bin_rhombically(xys)

        rxys = self.relative(xys)
        x = rxys["x"].values
        dx = self._ratio[0] * self._side
        x0 = (ijs.i.values + ijs.j.values) * dx / 2
        scaled_x = (x - x0) / dx

        correction = np.zeros(xys.shape[0])
        correction[scaled_x >= 0.5] = 1

        return Frame2D(i=(ijs.i.values + correction), j=(ijs.j.values + correction),
                       dtype=int, index=xys.index)

    def map_to_hexagonal(self, triangular_bins):
        """..."""
        ijs = triangular_bins
        N = ijs.shape[0]

        n = (ijs["j"] -  ijs["i"]).mod(3)

        is_mod1 = n == 1
        is_mod2 = n == 2

        i_correction_1 = np.zeros(N)
        i_correction_1[is_mod1] = 1
        correction_1 = Frame2D(i=i_correction_1, j=0,
                               index=ijs.index, dtype=int)

        j_correction_2 = np.zeros(N)
        j_correction_2[is_mod2] = 1
        correction_2 = Frame2D(i=0, j=j_correction_2,
                               index=ijs.index, dtype=int)

        return ijs + correction_1 + correction_2

    def bin_hexagonally(self, xys, use_columns_row_indexing=False):
        """..."""
        ijs = self.bin_rhombically(xys)

        rxys = self.relative(xys)
        x = rxys["x"].values
        dx = self._ratio[0] * self._side
        x0 = (ijs.i.values + ijs.j.values) * dx / 2
        scaled_x = (x - x0) / dx

        centers = (ijs["j"] -  ijs["i"]).mod(3) == 0
        correction = np.zeros(xys.shape[0])
        correction[scaled_x >= 0.5] = 1
        correction[~centers] = 0

        ijs = Frame2D(i=(ijs.i+correction), j=(ijs.j+correction),
                      dtype=int, index=ijs.index)

        hijs = self.map_to_hexagonal(triangular_bins=ijs)

        if not use_columns_row_indexing:
            return hijs
        return self.index_with_column_row(hijs)

    def index_with_column_row(self, hijs):
        """..."""
        d = (hijs["j"] - hijs["i"])
        n = d.mod(3)
        assert (n == 0).all()

        r = pd.Series(d/3, dtype=int)
        odd = r.mod(2) == 1

        cval = hijs["j"].values - 3 * r / 2
        cval[odd] = cval - 0.5
        c = pd.Series(cval, dtype=int)

        return Frame2D(col= c, row=r, dtype=int, index=hijs.index)

    def locate(self, bins):
        """Un bin the bins: (i, j) -> (x, y)
        """
        uvs = self._side * bins.rename(columns={"i": "u", "j": "v"})
        xys = self.reverse_transform(uvs)
        xys.index = pd.MultiIndex.from_frame(bins)
        return xys

    def annotate(self, gridpoints, using_column_row=False):
        """Annotate a gridpoints (x, y) in a dataframe indexed by their indices (i, j)
        on a triangular grid.
        """
        gridindex = gridpoints.index.to_frame()

        if not using_column_row:
            return gridindex.apply(lambda row: f"I{row.i};J{row.j}", axis=1)

        gridcolrows = self.index_with_column_row(gridindex)

        return gridcolrows.apply(lambda row: f"R{row.row};C{row.col}", axis=1)

    def locate_grid(self, bins):
        """..."""
        grid_points = self.locate(bins.drop_duplicates().reset_index(drop=True))
        return grid_points

    def plot_hextiles(self, positions, bins=None, graphic=None,
                      annotate=True, with_grid=True,
                      pointcolor=None, pointmarker="o", pointmarkersize=20,):
        """
        TODO: Annotate trigrid.
        """
        if annotate is True:
            annotate = "colrow"
        assert not annotate or annotate in ("hexgrid", "colrow")

        tiles = (self.bin_hexagonally(positions, use_columns_row_indexing=False)
                 if bins is None else bins)

        if graphic is None:
            raise ValueError("Need to specify figure and axes for plotting!")
        else:
            figure, axes = graphic

        if pointcolor is None:
            cr_tiles = self.index_with_column_row(tiles)
            even_col = np.array(cr_tiles.col.mod(2) == 0, dtype=bool)
            even_row = np.array(cr_tiles.row.mod(2) == 0, dtype=bool)

            palette = np.array(["red", "green", "blue", "orange"])
            colors_index = 2 * even_col + even_row
            colors = palette[colors_index]
        else:
            colors = pointcolor

        axes.scatter(positions["x"], positions["y"],
                    c=colors, marker=pointmarker, s=pointmarkersize)

        if with_grid:

            grid = self.locate_grid(tiles)

            axes.scatter(grid["x"], grid["y"], c="black", s=80)

        if annotate:

            annotate = self.annotate(grid, using_column_row=(annotate=="colrow"))
            for row in grid.assign(annotation=annotate).itertuples():
                axes.annotate(row.annotation, (row.x - 6, row.y + 5),
                              fontsize=20)

        return (figure, axes)



# SPDX-License-Identifier: Apache-2.0
import numpy
import pandas
from scipy.spatial.transform import Rotation
from .flatmap_utility import apply_flatmap_with_translation


# The below are just used internally in this package. Not quite redundant with the stuff in neuron_groups.defaults...
COLS_FLAT_XY = ["flat x", "flat y"]
COLS_XYZ = ["x", "y", "z"]


def flat_multi_index(coordinates3d, fm, orientations3d=None):
    coords_flat = apply_flatmap_with_translation(coordinates3d, orientations3d, fm)
    midx = pandas.MultiIndex.from_frame(pandas.DataFrame(coords_flat, columns=COLS_FLAT_XY))
    return midx


def flat_coordinate_frame(coordinates3d, fm, orientations3d=None, grouped=False):
    midx = flat_multi_index(coordinates3d, fm, orientations3d=orientations3d)
    coord_frame = pandas.DataFrame(coordinates3d, index=midx, columns=COLS_XYZ)
    if grouped:
        return coord_frame.groupby(COLS_FLAT_XY).apply(lambda x: x.values)
    return coord_frame


def pandas_flat_coordinate_frame(df_in, fm, columns_xyz=COLS_XYZ, columns_uvw=None, grouped=False):
    xyz = df_in[columns_xyz].values
    uvw = None
    if columns_uvw is not None:
        uvw = df_in[columns_uvw].values

    midx = flat_multi_index(xyz, fm, orientations3d=uvw)
    df_out = df_in.set_index(midx)
    if grouped:
        A = df_out[columns_xyz].groupby(COLS_FLAT_XY).apply(lambda x: x.values)
        return A, df_out
    return df_out


def voxel_flat_coordinate_frame(fm, in_voxel_indices=False, grouped=False):
    valid = numpy.all(fm.raw > -1, axis=-1)
    vxl_xyz = numpy.vstack(numpy.nonzero(valid)).transpose()
    vxl_flat = fm.raw[vxl_xyz[:, 0], vxl_xyz[:, 1], vxl_xyz[:, 2]]
    if not in_voxel_indices:
        vxl_xyz = vxl_xyz * fm.voxel_dimensions.reshape((1, -1)) + fm.offset.reshape((1, -1))
    vxl_frame = pandas.DataFrame(vxl_xyz, index=pandas.MultiIndex.from_tuples(map(tuple, vxl_flat),
                                                                              names=["f_x", "f_y"]),
                                 columns=["x", "y", "z"])
    if grouped:
        return vxl_frame.groupby(["f_x", "f_y"]).apply(lambda x: x.values)
    return vxl_frame


class Translation(object):
    def __init__(self, v):
        self._v = v

    def apply(self, other):
        return other + self._v

    def inv(self):
        return Translation(-self._v)


class Projection(object):
    def __init__(self, idx):
        self._idx = idx

    def apply(self, other):
        return other[:, self._idx]


class GeneralLinearTransform(object):
    def __init__(self, M):
        self._M = M

    def apply(self, other):
        return numpy.dot(self._M, other.transpose()).transpose()

    def inv(self):
        return GeneralLinearTransform(self._M.transpose())

    def expand(self):
        Mout = numpy.array([
            [self._M[0, 0], 0, self._M[0, 1]],
            [0, 1, 0],
            [self._M[1, 0], 0, self._M[1, 1]]
        ])
        return GeneralLinearTransform(Mout) #Rotation.from_matrix(Mout)


class Combination(object):
    def __init__(self, one, two):
        self._one = one
        self._two = two

    def apply(self, other):
        return self._two.apply(self._one.apply(other))

    def inv(self):
        return Combination(self._two.inv(), self._one.inv())


def flatmap_pixel_gradient(fm_or_frame):
    from .flatmap_utility import colored_points_to_image
    if not isinstance(fm_or_frame, pandas.DataFrame):
        fm_or_frame = voxel_flat_coordinate_frame(fm_or_frame)
    per_pixel = fm_or_frame.groupby(["f_x", "f_y"])
    per_pixel_center = per_pixel.apply(lambda x: numpy.mean(x.values, axis=0))

    pxl_center_vol = colored_points_to_image(numpy.vstack(per_pixel_center.index.values),
                                             numpy.vstack(per_pixel_center.values))
    # Gradients thereof: When we go one step in the flat space, this is how many steps we going in global coordinates
    dx_dfx, dx_dfy = numpy.gradient(pxl_center_vol[:, :, 0])
    dy_dfx, dy_dfy = numpy.gradient(pxl_center_vol[:, :, 1])
    dz_dfx, dz_dfy = numpy.gradient(pxl_center_vol[:, :, 2])

    dfx = numpy.dstack([dx_dfx, dy_dfx, dz_dfx])
    dfy = numpy.dstack([dx_dfy, dy_dfy, dz_dfy])
    return dfx, dfy


def _find_rotation_(v_x, v_y):
    if numpy.any(numpy.isnan(v_x)):
        if numpy.any(numpy.isnan(v_y)):
            return GeneralLinearTransform(numpy.identity(2)), 2.0
        vv = numpy.hstack([v_y, [[0]]])
        vtgt = numpy.array([[0, 1, 0]])
    elif numpy.any(numpy.isnan(v_y)):
        vv = numpy.hstack([v_x, [[0]]])
        vtgt = numpy.array([[1, 0, 0]])
    else:
        vv = numpy.hstack([numpy.vstack([v_x, v_y]), [[0], [0]]])
        vtgt = numpy.array([[1, 0, 0], [0, 1, 0]])
    vv = vv / numpy.linalg.norm(vv, axis=1, keepdims=True)
    res = Rotation.align_vectors(vtgt, vv)
    M = res[0].as_matrix()
    return GeneralLinearTransform(M[:2, :2]), res[1]



# SPDX-License-Identifier: Apache-2.0
# Functionality for adding additional columns to the DataFrame holding neuron information
# That is: information that is not already easily available using Circuit.cells.get
import pandas
import numpy

from ...flatmapping import supersampled_locations, apply_flatmap_with_translation
from .defaults import XYZ, UVW, DEFAULT_PIXEL_SZ, SS_COORDINATES, FLAT_COORDINATES


def supersampled_locations_wrapper(df_in, circ=None, fm=None, orient=None, pixel_sz=DEFAULT_PIXEL_SZ):
    """
    Wraps the lookup of supersampled flat locations and depth of fibers and neurons. Can be used for
    both fiber and neuron locations.
    Input:
    df_in (pandas.DataFrame): Holds already loaded information on neurons or fibers. Must contain at
    least three columns with x, y and z coordinates. If it also contains u, v, w columns, they may be
    used to translate x, y, z locations into the flat mapped volume (required for fibers).
    fm (str, optional): Path to flatmap file to use
    orient (str, optional): Path to orientation volume to use
    circ (bluepysnap.Circuit, optional): Can be provided instead of fm and orient. In that case flatmap and
    orientation are loaded from Circuit.atlas
    pixel_sz (float): Assumed approximate size in um of a single flatmap pixel. If set to None, an 
    approximate value will be estimated

    Returns:
    a copy of df_in with two additional columns holding the supersampled flat coordinates and depth
    """
    assert numpy.all([col in df_in for col in XYZ]), "Must load base x, y, z coordinates first"
    columns_uvw = None
    if numpy.all([col in df_in for col in UVW]):
        columns_uvw = UVW
    return supersampled_locations(df_in, XYZ, columns_uvw=columns_uvw, circ=circ, fm=fm, orient=orient,
                                  pixel_sz=pixel_sz, columns_out=SS_COORDINATES, include_depth=True)


def flat_locations(df_in, circ, fm=None):
    """
    Adds simple flatmapped locations to a DataFrame holding neuron information.
    Input:
    df_in (pandas.DataFrame): Holds already loaded information on neurons or fibers. Must contain at
    least three columns with x, y and z coordinates. If it also contains u, v, w columns, they may be
    used to translate x, y, z locations into the flat mapped volume (required for fibers).
    fm (str, optional): Path to flatmap file to use
    circ (bluepysnap.Circuit): Can be provided instead of fm. In that case flatmap is loaded from
    Circuit.atlas. If fm is provided, then circ can be None.
    """
    assert numpy.all([col in df_in for col in XYZ]), "Must load base x, y, z coordinates first"
    if fm is None:
        from .sonata_extensions import load_atlas_data
        fm = load_atlas_data(circ, "flatmap")
    xyz = df_in[XYZ].values
    uvw = None
    if numpy.all([x in df_in for x in UVW]):
        uvw = df_in[UVW].values
    flat_coords = apply_flatmap_with_translation(xyz, uvw, fm)
    return pandas.DataFrame(flat_coords,
                            columns=FLAT_COORDINATES,
                            index=df_in.index)


AVAILABLE_EXTRAS = {
    flat_locations: FLAT_COORDINATES,
    supersampled_locations_wrapper: SS_COORDINATES
}


def add_extra_properties(df_in, circ, lst_properties, fm=None):
    """
    Adds additional properties (like flat loactions) to a DataFrame holding neuron or fiber information. 
    List of available properties:
    flat_x: Flat mapped x coordinate (pixel resolution)
    flat_y: Flat mapped y coordinate (pixel resolution)
    ss_flat_x: Supersampled flat mapped x coordinate (approximately in um)
    ss_flat_y: Supersampled flat mapped y coordinate (approximately in um)
    depth: Approximate cortical depth of a neuron in um

    Input:
    df_in (pandas.DataFrame): Holds already loaded information on neurons or fibers. Must contain at
    least three columns with x, y and z coordinates. If it also contains u, v, w columns, they may be
    used to translate x, y, z locations into the flat mapped volume (required for fibers).
    lst_properties (list-like): List of properties to add. For available properties: see above.
    circ (bluepysnap.Circuit): Circuit that the neurons or fibers originate from. Must have a "orientation"
    atlas in Circuit.atlas!
    fm (str, optional): Path to flat map file to use. If not provided, then circ must have a "flatmap"
    atlas in Circuit.atlas!

    Returns:
    a copy of df_in with additional columns holding the specified additional properties.
    """
    lst_properties = list(lst_properties)
    for extra_fun, extra_props in AVAILABLE_EXTRAS.items():
        props = []
        for p in extra_props:
            if p in lst_properties:
                props.append(lst_properties.pop(lst_properties.index(p)))
        if len(props) > 0:
            new_df = extra_fun(df_in, circ, fm=fm)
            df_in = pandas.concat([df_in, new_df[props]], axis=1, copy=False)
    if len(lst_properties) > 0:
        raise ValueError("Unknown properties: {0}".format(lst_properties))
    return df_in



# SPDX-License-Identifier: Apache-2.0
import pandas
import numpy


def expand_partition(Mc, label_use="new_partition", label_original="_idxx_in_original"):
    L = Mc.vertices[label_original].apply(len).sum()
    lbls = -numpy.ones(L, dtype=int)
    for lbl, idxx in Mc.vertices.groupby(label_use)["_idxx_in_original"].apply(numpy.hstack).items():
        lbls[idxx] = lbl
    return lbls


def ren_eel(M, clust_funcs, clust_func_update, kmax=None, modularity_kwargs={}):
    """
    https://www.nature.com/articles/s41598-019-50739-3
    """
    if isinstance(clust_funcs, list):
        clust_funcs = dict([("base_partition_{0}".format(i), _func)
                            for i, _func in enumerate(clust_funcs)])
    if kmax is None: kmax = len(clust_funcs)
    assert kmax >= len(clust_funcs)

    func_names = list(clust_funcs.keys())
    for k, _func in clust_funcs.items():
        M.add_vertex_property(k, _func(M.matrix))
    
    scoreboard = pandas.Series([
        M.modularity(k, **modularity_kwargs) for k in func_names
    ], index=func_names)
    cand_count = 0

    while(True):
        Mc = M.condense(func_names)
        Mc.add_vertex_property("new_partition", clust_func_update(Mc.matrix))
        cand_str = "candidate_{0}".format(cand_count)
        cand_count += 1
        M.add_vertex_property(cand_str, expand_partition(Mc))

        if numpy.any([(M.vertices[_k] == M.vertices[cand_str]).all()
                    for _k in list(scoreboard.index)]):
            scoreboard.pop(scoreboard.idxmin())
        else:                
            cand_score = M.modularity(cand_str, **modularity_kwargs)
            if cand_score > scoreboard.min():
                if len(scoreboard) >= kmax:
                    scoreboard.pop(scoreboard.idxmin())
                scoreboard[cand_str] = cand_score
            else:
                scoreboard.pop(scoreboard.idxmin())
            
        print("""
        Iteration {0}:
            k = {1}
            min_score = {2}
            mean_score = {3}
            max_score = {4}

        """.format(
            cand_count - 1,
            scoreboard.min(),
            scoreboard.mean(),
            scoreboard.max()
        ))
        if len(scoreboard) == 1:
            break
    return scoreboard

    
                        




# SPDX-License-Identifier: Apache-2.0
# For loading connectivity matrices
import h5py
import numpy
import tqdm
import pandas
import os
from scipy import sparse

from .neuron_groups.defaults import GID
from .sonata_helpers import LOCAL_CONNECTOME, find_sonata_connectome, get_connectome_shape

STR_VOID = "VOID"


def full_connection_matrix(sonata_fn, n_neurons=None, population="default",
                           edge_property=None, agg_func=None, shape=None, chunk=50000000):
    """
    Returns the full connection matrix encoded in a sonata h5 file.
    Input:
    sonata_fn (str): Path to the sonata h5 file.
    n_neurons (optional): Number of neurons in the connectome. If not provided, it will be estimated
                          from the connectivity info, but unconnected neurons may get ignored!
    population (str): Sonata population to work with.
    edge_property (str, optional): Name of a synapse property to look up. Must exist in the connectome.
    If not provided, a boolean matrix is returned
    agg_func (list, optional): Name of aggregation function to apply to the property of synapses belonging to the same connection.
    Must be provided if edge_property is provided!
    chunk (optional): Number of connections to read at the same time. Larger values
    will run generally faster, but with fewer updates of the progress bar.

    Returns:
    scipy.sparse matrix of connectivity
    """
    if edge_property is not None:
        assert agg_func is not None, "Must also provide a list of functions to aggregate synapses belonging to the same connection!"
        return _full_connection_property(sonata_fn, edge_property, agg_func, n_neurons=n_neurons,
                                         population=population, shape=shape, chunk=chunk)
    h5 = h5py.File(sonata_fn, "r")['edges/%s' % population]
    if n_neurons is not None and shape is None:
        shape = (n_neurons, n_neurons)

    dset_sz = h5['source_node_id'].shape[0]
    A = numpy.zeros(dset_sz, dtype=int)
    B = numpy.zeros(dset_sz, dtype=int)
    splits = numpy.arange(0, dset_sz + chunk, chunk)
    for splt_fr, splt_to in tqdm.tqdm(zip(splits[:-1], splits[1:]), total=len(splits) - 1):
        A[splt_fr:splt_to] = h5['source_node_id'][splt_fr:splt_to]
        B[splt_fr:splt_to] = h5['target_node_id'][splt_fr:splt_to]
    M = sparse.coo_matrix((numpy.ones_like(A, dtype=bool), (A, B)), shape=shape)
    return M.tocsr()


def _full_connection_property(sonata_fn, edge_property, agg_func, n_neurons=None, population="default",
                              shape=None, chunk=50000000):
    """
    Returns the full connection matrix encoded in a sonata h5 file. Instead of just a binary matrix, it looks up
    and assigns structural properties to the connections
    Input:
    sonata_fn (str): Path to the sonata h5 file.
    edge_property (str): Name of the property to look up. Must exist in the connectome
    agg_func (list): Name of aggregation function to apply to the property of synapses belonging to the same connection
    n_neurons (optional): Number of neurons in the connectome. If not provided, it will be estimated
                          from the connectivity info, but unconnected neurons may get ignored!
    population (str): Sonata population to work with.
    chunk (optional): Number of connections to read at the same time. Larger values
    will run generally faster, but with fewer updates of the progress bar.

    Returns:
    scipy.sparse matrix of connectivity
    """
    if not isinstance(agg_func, list):
        raise NotImplementedError("Currently only implemented for list-of-functions")
    h5 = h5py.File(sonata_fn, "r")['edges/%s' % population]
    if n_neurons is not None and shape is None:
        shape = (n_neurons, n_neurons)

    dset_sz = h5['source_node_id'].shape[0]
    row_indices = []
    col_indices = []
    out_data = dict([(afunc, []) for afunc in agg_func])

    splits = numpy.arange(0, dset_sz + chunk, chunk)
    for splt_fr, splt_to in tqdm.tqdm(zip(splits[:-1], splits[1:]), total=len(splits) - 1):
        A = h5['source_node_id'][splt_fr:splt_to]
        B = h5['target_node_id'][splt_fr:splt_to]
        data = h5['0'][edge_property][splt_fr:splt_to]
        S = pandas.Series(data, index=pandas.MultiIndex.from_arrays([A, B])).groupby(level=[0, 1]).agg(agg_func)
        Si = S.index.to_frame()
        row_indices.extend(Si[0].values); col_indices.extend(Si[1].values)
        for afunc in agg_func:
            out_data[afunc].extend(S[afunc].values)

    M = dict([
        (afunc,
         sparse.coo_matrix((out_data[afunc], (row_indices, col_indices)), shape=shape).tocsr())
         for afunc in agg_func
    ])
    return M


def _connection_property_for_gids(sonata_fn, gids, gids_post, population, edge_property, agg_func):
    """
    Returns the connection matrix encoded in a sonata h5 file for a subset of neurons.
    Input:
    sonata_fn (str): Path to the sonata h5 file.
    gids: List of neuron gids to get the connectivity for.
    gids_post (optional): If given, then connectivity FROM gids TO gids_post will be returned. Else: FROM gids TO gids.
    population (str): Sonata population to work with.
    edge_property (str): Name of the synapse property to report as the value of a connection
    agg_func: if a callable: A function to aggregate multiple synapse properties into a single,
                             scalar connection property (has to work with pandas' `.apply()`)
              if a list: A list of the above described aggregation functions (has to work with pandas' `.agg()`)

    Returns:
    scipy.sparse matrix of connectivity (or a dict of those if a list is passed as `agg_func`)
    """
    h5 = h5py.File(sonata_fn, "r")['edges/%s' % population]

    idx = numpy.array(gids)
    if gids_post is None:
        gids_post = gids
    idx_post = numpy.array(gids_post)
    rv_index = pandas.Series(range(len(idx)), index=idx)
    N = len(gids)
    M = len(gids_post)

    indices = []
    indptr = [0]
    if not isinstance(agg_func, list):
        data = []
        for id_post in tqdm.tqdm(idx_post):
            ids_pre = []
            data_pre = []
            ranges = h5['indices']['target_to_source']['node_id_to_ranges'][id_post, :]
            for block in h5['indices']['target_to_source']['range_to_edge_id'][ranges[0]:ranges[1], :]:
                ids_pre.append(h5['source_node_id'][block[0]:block[1]])
                data_pre.append(h5['0'][edge_property][block[0]:block[1]])

            if len(ids_pre) > 0:
                ids_pre = numpy.hstack(ids_pre)
                row_ids = rv_index[rv_index.index.intersection(ids_pre)].values
                indices.extend(row_ids)
                data_pre = pandas.Series(numpy.hstack(data_pre), index=ids_pre)
                data_pre = data_pre[data_pre.index.intersection(idx)]
                data.extend(data_pre.groupby(level=0, group_keys=False).apply(agg_func).values)

            indptr.append(len(indices))
        mat = sparse.csc_matrix((data, indices, indptr), shape=(N, M))
        return mat
    else:
        data = {agg_f: [] for agg_f in agg_func}
        for id_post in tqdm.tqdm(idx_post):
            ids_pre = []
            data_pre = []
            ranges = h5['indices']['target_to_source']['node_id_to_ranges'][id_post, :]
            for block in h5['indices']['target_to_source']['range_to_edge_id'][ranges[0]:ranges[1], :]:
                ids_pre.append(h5['source_node_id'][block[0]:block[1]])
                data_pre.append(h5['0'][edge_property][block[0]:block[1]])

            if len(ids_pre) > 0:
                ids_pre = numpy.hstack(ids_pre)
                row_ids = rv_index[rv_index.index.intersection(ids_pre)].values
                indices.extend(row_ids)
                data_pre = pandas.Series(numpy.hstack(data_pre), index=ids_pre)
                data_pre = data_pre[data_pre.index.intersection(idx)]
                # here is the main difference from the one below
                # (this is needed to be in sync. for `.io.synapse_report/aggregate_data()`)
                res = data_pre.groupby(level=0, group_keys=False).agg(agg_func)
                for agg_f in agg_func:
                    data[agg_f].extend(res[agg_f].to_numpy())

            indptr.append(len(indices))
        mats = {agg_f: sparse.csc_matrix((data[agg_f], indices, indptr), shape=(N, M)) for agg_f in agg_func}
        return mats


def connection_matrix_for_gids(sonata_fn, gids, gids_post=None, population="default",
                               edge_property=None, agg_func=None, load_full=False, **kwargs):
    """
    Returns the connection matrix encoded in a sonata h5 file for a subset of neurons.
    Input:
    sonata_fn (str): Path to the sonata h5 file.
    gids: List of neuron gids to get the connectivity for.
    gids_post (optional): If given, then connectivity FROM gids TO gids_post will be returned. Else: FROM gids TO gids.
    population (str): Sonata population to work with.
    edge_property (optional, str): Name of the synapse property to report as the value of a connection
                                   If not provided, "True" will be reported for all connections.
                                   If provided, _must_ also provide `agg_func`.
    agg_func: if a callable: A function to aggregate multiple synapse properties into a single,
                             scalar connection property (has to work with pandas' `.apply()`)
              if a list: A list of the above described aggregation functions (has to work with pandas' `.agg()`)

    Returns:
    scipy.sparse matrix of connectivity (or a dict of those if a list is passed as `agg_func`)
    """
    if gids_post is None:
        gids_post = gids
    if load_full:
        M = full_connection_matrix(sonata_fn, population=population, edge_property=edge_property,
                                   agg_func=agg_func, **kwargs)
        if isinstance(M, dict): return dict([(k, v.tocsr()[numpy.ix_(gids, gids_post)])
                                             for k, v in M.items()])
        return M.tocsr()[numpy.ix_(gids, gids_post)]
    if edge_property is not None:
        assert agg_func is not None, "When looking up connection properties, must provide an agg_func, such as 'mean'"
        return _connection_property_for_gids(sonata_fn, gids, gids_post, population, edge_property, agg_func)
    
    h5 = h5py.File(sonata_fn, "r")['edges/%s' % population]
    idx = numpy.array(gids)
    idx_post = numpy.array(gids_post)

    rv_index = pandas.Series(range(len(idx)), index=idx)
    N = len(gids)
    M = len(gids_post)

    indices = []
    indptr = [0]
    for id_post in tqdm.tqdm(idx_post):
        ids_pre = []
        ranges = h5['indices']['target_to_source']['node_id_to_ranges'][id_post, :]
        for block in h5['indices']['target_to_source']['range_to_edge_id'][ranges[0]:ranges[1], :]:
            ids_pre.append(h5['source_node_id'][block[0]:block[1]])
        if len(ids_pre) > 0:
            row_ids = rv_index[rv_index.index.intersection(numpy.hstack(ids_pre))].values
            indices.extend(row_ids)
        indptr.append(len(indices))
    mat = sparse.csc_matrix((numpy.ones(len(indices), dtype=bool), indices, indptr), shape=(N, M))
    return mat


def circuit_connection_matrix(circ, connectome=LOCAL_CONNECTOME, for_gids=None, for_gids_post=None,
                              edge_population=None, node_population=None,
                              edge_property=None, agg_func=None, chunk=50000000, load_full=False):
    """
    Returns a structural connection matrix, either for an entire circuit, or a subset of neurons.
    For either local connectivity or any projection. 
    Input:
    circ (bluepysnap.Circuit)
    connectome (str): Which connectome to return. Can be any EdgePopulation in circ.edges. If not provided,
                      it will be heuristically guessed using the value of node_population. The first EdgePopulation
                      that has the specified node_population as both source and target will be used. If 
                      node_population is also not provided, then the largest non-virtual population will be used.
    for_gids: List of neuron gids to get the connectivity for.
    for_gids_post (optional): If given, then connectivity FROM for_gids TO for_gids_post will be returned.
                              Else: FROM for_gids TO for_gids.
                              NOTE: for_gids_post will be ignored if for_gids is not provided!
                              NOTE: Can be used to get the matrix of external innervation!
                              For that purpose, provide the gids of innervating fibers as for_gids,
                              the gids of circuit neurons as for_gids_post and the name or the projection as connectome.
    edge_population (str, optional): Sonata edge population name. Usually not required to be specified.
    node_population (str, optional): Name of a node population. Only used to look up local connectomes when
                              connectome="local" is used.
    edge_property (optional, str): Name of the synapse property to report as the value of a connection
                                   If not provided, "True" will be reported for all connections.
                                   If provided, _must_ also provide `agg_func`.
    agg_func: if a callable: A function to aggregate multiple synapse properties into a single,
                             scalar connection property (has to work with pandas' `.apply()`)
              if a list: A list of the above described aggregation functions (has to work with pandas' `.agg()`)
    chunk (optional): Number of connections to read at the same time. Larger values will run generally faster,
                      but with fewer updates of the progress bar.

    Returns:
    scipy.sparse matrix of connectivity (or a dict of those if a list is passed as `agg_func`)
    Note: By default, returns binary connectivity, i.e. only the presence or absence of at least one connection between
    nodes. If you want to count connections, use the edge_property=... with any valid edge property and 
    agg_func=len
    """
    if connectome == LOCAL_CONNECTOME: 
        from .sonata_helpers import local_connectome_for, nonvirtual_node_population
        if node_population is None: node_population = nonvirtual_node_population(circ)
        connectome = local_connectome_for(circ, node_population)
    conn_file = find_sonata_connectome(circ, connectome)
    shape = get_connectome_shape(circ, connectome)
    if edge_population is None:
        edge_population = connectome
    if for_gids is None:
        return full_connection_matrix(conn_file, edge_property=edge_property, agg_func=agg_func, shape=shape,
                                      population=edge_population, chunk=chunk)
    return connection_matrix_for_gids(conn_file, for_gids, gids_post=for_gids_post, population=edge_population,
                                      edge_property=edge_property, agg_func=agg_func, load_full=load_full, chunk=chunk)


def circuit_node_set_matrix(circ, for_node_set, for_node_set_post=None):
    """
    Returns a structural connection matrix within or between defined node sets. That is, unlike 
    circuit_connection_matrix this function can aggregate over multiple edge_populations and
    node_populations.
    """
    from .sonata_helpers import resolve_node_set

    if not isinstance(for_node_set, pandas.DataFrame):
        node_set_pre = resolve_node_set(circ, for_node_set).reset_index().set_index("population")
    else: node_set_pre = for_node_set.reset_index().set_index("population")
    if for_node_set_post is None: node_set_post = node_set_pre
    elif not isinstance(for_node_set_post, pandas.DataFrame):
        node_set_post = resolve_node_set(circ, for_node_set_post).reset_index().set_index("population")
    else: node_set_post = for_node_set_post.reset_index().set_index("population")

    row = []; col = []; data = []
    for edge_name, edges in circ.edges.items():
        rel_pre = edges.source.name in node_set_pre.index
        rel_post = edges.target.name in node_set_post.index
        if rel_pre and rel_post:
            tM = circuit_connection_matrix(circ, connectome=edge_name,
                                           for_gids=node_set_pre["node_ids"][edges.source.name],
                                           for_gids_post=node_set_post["node_ids"][edges.target.name]).tocoo()
            tgt_ids = node_set_post["index"][edges.target.name]
            src_ids = node_set_pre["index"][edges.source.name]
            
            row.extend(src_ids.iloc[tM.row])
            col.extend(tgt_ids.iloc[tM.col])
            data.extend(tM.data)
    M_out = sparse.coo_matrix((data, (row, col)), shape=(len(node_set_pre), len(node_set_post)))
    node_set_pre = node_set_pre.reset_index().set_index("index")
    node_set_post = node_set_post.reset_index().set_index("index")
    return M_out, node_set_pre, node_set_post


def circuit_group_matrices(circ, neuron_groups, connectome=LOCAL_CONNECTOME, extract_full=False,
                           column_gid=GID, **kwargs):
    """
    Returns matrices of the structural connectivity within specified groups of neurons.
    For any EdgePopulation (specified as connectome=...).

    Note: This function strongly assumes that the source and target node population of the specified
    connectome match the nodes specified in neuron_groups. No check is performed!

    Input:
    circ (bluepysnap.Circuit)
    neuron_groups (pandas.DataFrame): Frame of neuron grouping info. 
    See conntility.circuit_models.neuron_groups for information how a group is defined.
    connectome (str, default: "local"): Which connectome to return. Can be either "local", returning the 
    touch connectome or the name of any projection defined in the CircuitConfig.
    extract_full (bool, default: False): If set to True, this will first extract the _complete_ connection
    matrix between all neurons, then look up the relevant parts of that huge matrix. This can be faster,
    if the neuron_groups are a partition of all neurons, or close to it. But it uses much more memory.
    column_gid (str, default: "gid"): Name of the column of neuron_groups that holds the gid of a neuron.

    Returns:
    pandas.DataFrame of scipy.sparse matrices of connectivity. Index of the frame is identical to the index
    of neuron_groups.
    """
    if isinstance(neuron_groups, pandas.DataFrame):
        neuron_groups = neuron_groups[column_gid]
    neuron_groups = neuron_groups.groupby(neuron_groups.index.names)
    if not extract_full:
        matrices = neuron_groups.apply(lambda grp: circuit_connection_matrix(circ, connectome=connectome,
                                                                             for_gids=grp.values, **kwargs))
    else:
        full_matrix = circuit_connection_matrix(circ, connectome=connectome, **kwargs)
        matrices = neuron_groups.apply(lambda grp: full_matrix[numpy.ix_(grp.values, grp.values)])
    matrices = matrices[matrices.apply(lambda x: isinstance(x, sparse.spmatrix))]
    return matrices


def circuit_cross_group_matrices(circ, neuron_groups_pre, neuron_groups_post, connectome=LOCAL_CONNECTOME,
                                 extract_full=False, column_gid=GID, **kwargs):
    """
    Returns the structural connectivity between (and within) specified groups of neurons.
    That is, a number of matrices of structural connectivity between neurons in group A and B. 
    This can be thought of as one big matrix, broken up into sub-matrices by group.
    For any EdgePopulation (specified as connectome=...).

    Note: This function strongly assumes that the source and target of the EdgePopulation match the 
    nodes given in neuron_groups_pre and neuron_groups_post. No check is performed!

    Input:
    circ (bluepy.Circuit)
    neuron_groups_pre (pandas.DataFrame): Frame of neuron grouping info. 
    See conntility.circuit_models.neuron_groups for information how a group is defined. 
    These groups will be the groups _sending_ the connections that are considered.
    neuron_groups_post (pandas.DataFrame): Frame of neuron grouping info. 
    These groups will be the groups _receiving_ the connections that are considered. Can be the same as 
    neuron_groups_pre.
    connectome (str, default: "local"): Which connectome to return. Can be any EdgePopulation in circ.edges.
    If "local", a local recurrent connectome will be heuristically guessed. In that case, it is best to
    also specify the name of a node population using node_population=...
    The way a "local" connectome is guessed from node_population=... is documented in circuit_connection_matrix()
    extract_full (bool, default: False): If set to True, this will first extract the _complete_ connection
    matrix between all neurons, then look up the relevant parts of that huge matrix. This can be faster,
    if the neuron_groups are a partition of all neurons, or close to it. But it uses much more memory.
    column_gid (str, default: "gid"): Name of the column of neuron_groups pre/post that holds the gid of a neuron.

    Returns:
    pandas.DataFrame of scipy.sparse matrices of connectivity. Index of the frame is identical to the index
    of neuron_groups_pre. Columns are indentical to the index of neuron_groups_post.
    """
    if extract_full:
        full_matrix = circuit_connection_matrix(circ, connectome=connectome, **kwargs)

        def prepare_indexing(df_pre):
            def index_submat(df_post):
                return full_matrix[numpy.ix_(df_pre[column_gid].values, df_post[column_gid].values)]
            return index_submat

        res = neuron_groups_pre.groupby(neuron_groups_pre.index.names).apply(
            lambda df_pre:
            neuron_groups_post.groupby(neuron_groups_post.index.names).apply(prepare_indexing(df_pre))
        )
        return res

    def prepare_con_mat(df_pre):
        def execute_con_mat(df_post):
            if len(df_pre) == 0 or len(df_post) == 0:
                _shape = (len(df_pre), len(df_post))
                return sparse.csc_matrix(numpy.empty(shape=_shape, dtype=bool))
            return circuit_connection_matrix(circ, for_gids=df_pre[column_gid].values,
                                             for_gids_post=df_post[column_gid].values,
                                             connectome=connectome, **kwargs)

        return execute_con_mat

    res = neuron_groups_pre.groupby(neuron_groups_pre.index.names).apply(
        lambda df_pre:
        neuron_groups_post.groupby(neuron_groups_post.index.names).apply(prepare_con_mat(df_pre))
    )
    return res


def _make_node_lookup(circ, neuron_groups, column_gid, fill_unused_gids=True):
    from .neuron_groups import flip
    node_lookup = flip(neuron_groups, index=column_gid, contract_values=True, categorical=~fill_unused_gids)
    if fill_unused_gids:
        node_ids = circ.nodes.ids()
        all_gids = node_ids.index.to_frame()[GID].values
        missing_gids = numpy.setdiff1d(all_gids, node_lookup.index)
        full_lookup = pandas.concat([node_lookup,
                                     pandas.Series([STR_VOID] * len(missing_gids),
                                                   index=missing_gids)], axis=0)
        node_lookup = pandas.Series(pandas.Categorical(full_lookup), index=full_lookup.index, name=node_lookup.name)
    return node_lookup


def connection_matrix_between_groups_partition(sonata_fn, node_lookup, population, chunk=50000000):
    # TODO: If the user accidently provides a "neuron_groups" instead of "node_lookup" input give helpful message
    # TODO: Evaluate if it is necessary to fill node_lookup for unused gids with STR_VOID
    """
    Don't use this. Use circuit_matrix_between_groups
    """
    h5 = h5py.File(sonata_fn, "r")['edges/{0}'.format(population)]  # TODO: close file!

    dset_sz = h5['source_node_id'].shape[0]
    splits = numpy.arange(0, dset_sz + chunk, chunk)

    midxx = pandas.MultiIndex.from_tuples([], names=["Source node", "Target node"])
    counts = pandas.Series([], index=midxx, dtype=int)

    for splt_fr, splt_to in tqdm.tqdm(zip(splits[:-1], splits[1:]), desc="Counting...", total=len(splits) - 1):
        son_idx_fr = h5['source_node_id'][splt_fr:splt_to]
        son_idx_to = h5['target_node_id'][splt_fr:splt_to]
        reg_fr = node_lookup[son_idx_fr]
        reg_to = node_lookup[son_idx_to]
        new_counts = pandas.DataFrame({"Source node": reg_fr.values,
                                       "Target node": reg_to.values}).value_counts()
        counts = counts.add(new_counts, fill_value=0)

    for lvl, nm in zip(counts.index.levels, counts.index.names):
        if STR_VOID in lvl:
            counts = counts.drop(STR_VOID, level=nm)
    return counts


def _afferent_gids(h5, post_gid):
    rnge = h5["indices"]["target_to_source"]["node_id_to_ranges"][post_gid]
    if rnge[1] == rnge[0]:
        return numpy.array([])
    son_idx_fr = [h5["source_node_id"][r[0]:r[1]]
                  for r in h5["indices"]["target_to_source"]["range_to_edge_id"][rnge[0]:rnge[1]]]
    son_idx_fr = numpy.hstack(son_idx_fr)
    return son_idx_fr


def connection_matrix_between_groups_partial(sonata_fn, node_lookup, population="default", **kwargs):
    # TODO: If the user accidently provides a "neuron_groups" instead of "node_lookup" input give helpful message
    """
    Don't use this. Use circuit_matrix_between_groups
    """
    node_lookup = node_lookup[node_lookup != STR_VOID]
    gids_per_node = node_lookup.to_frame().groupby(node_lookup.name).apply(lambda x: x.index.values)

    lst_node_to = []
    lst_counts_from = []
    with h5py.File(sonata_fn, "r") as h5_file:
        h5 = h5_file['edges/{0}'.format(population)]
        for node_to, lst_post_gids in tqdm.tqdm(gids_per_node.items(), total=len(gids_per_node)):
            lst_pre_gids = [_afferent_gids(h5, post_gid) for post_gid in lst_post_gids]
            lst_pre_gids = numpy.hstack(lst_pre_gids)
            lst_pre_gids = lst_pre_gids[numpy.in1d(lst_pre_gids, node_lookup.index)]
            node_from = node_lookup[lst_pre_gids]
            counts_from = node_from.value_counts()
            counts_from.index.name = "Source node"
            lst_node_to.append(node_to)
            lst_counts_from.append(counts_from)
    counts = pandas.concat(lst_counts_from, keys=lst_node_to, names=["Target node"])

    return counts


def circuit_matrix_between_groups(circ, neuron_groups, connectome,
                                  edge_population=None, extract_full=False, column_gid=GID):
    """
    Returns the number of structural connections between (and within) specified groups of neurons.
    That is, single matrix of the _number_ of connections between group A and B.  
    This can be thought of as a connectome with reduced resolution, as it is similar to
    a voxelized connectome. In fact, if the neuron_groups are based on binned x, y, z coordinates,
    this essentially returns a voxelized connectome!
    For any EdgePopulation in circ.edges.

    Note: This function strongly assumes that the source and target of the specified connectome match the
    neurons given in neuron_groups. It is up to the user to ensure that!
    Note: If there are multiple connection (synapses) between a pair of nodes, they will be counted multiple
    times!

    Input:
    circ (bluepysnap.Circuit)
    neuron_groups (pandas.DataFrame): Frame of neuron grouping info. 
    See conntility.circuit_models.neuron_groups for information how a group is defined.
    connectome (str): Which connectome to return. Must be in circ.edges.
    edge_population (str, optional): Name of the edge population in the connectome. Usually not needed to be specified.
    extract_full (bool, default: False): If set to True, this will first extract the _complete_ connection
    matrix between all neurons, then look up and sum the relevant parts of that huge matrix. This can be faster,
    if the neuron_groups are a partition of all neurons, or close to it. But it uses much more memory.
    column_gid (str, default: "node_ids"): Name of the column of neuron_groups pre/post that holds the gid of a neuron.

    Returns:
    pandas.DataFrame of connection counts. Index of the frame will be a MultiIndex with the columns
    "Source node" and "Target node". Values of these entries will be strings generated from the 
    MultiIndex of neuron_groups. This is a bit awkward and will be improved in the future.
    Use .stack and .values to turn this output into a classic connection matrix (2d numpy array).

    Note: You could use this to count structural innervation strength from external innervation. But you would
    have to concatenate a group definition for the innervating fibers and a group definition for the circuit
    neurons. 
    """
    # TODO: Support "local" connectome?
    # TODO: Support non-recurrent connectome!
    conn_file = find_sonata_connectome(circ, connectome, assert_is_recurrent=True)
    if edge_population is None:
        edge_population = connectome

    if extract_full:
        node_lookup = _make_node_lookup(circ, neuron_groups, column_gid)
        return connection_matrix_between_groups_partition(conn_file, node_lookup, population=edge_population)
    else:
        node_lookup = _make_node_lookup(circ, neuron_groups, column_gid, fill_unused_gids=False)
        return connection_matrix_between_groups_partial(conn_file, node_lookup, population=edge_population)



# SPDX-License-Identifier: Apache-2.0
# Functionality for defining groups of neurons according to their properties
import numpy
import pandas

from .tessellate import TriTille
from .defaults import GID

def group_by_properties(df_in, lst_props, prefix="idx-", replace=True):
    """
    Defines a group of neurons by their values of a list of specified properties.
    That is, each unique combination of values defines a separate group, so don't
    use this for float-valued properties.
    Groups are defined by the index of a DataFrame holding their properties.

    Input:
    df_in (pandas.DataFrame): Holds information on neurons or innervating fibers.
    One row per neuron / fiber, one column per property.
    lst_props (list-like): List of columns to use for grouping.
    prefix (str): A prefix to preprepend to the names of the used columns/properties
    for their use in the index. This is to avoid having columns in the frame and index
    with the same name, which can break .groupby statements.
    replace (bool, default=True): Whether to replace the existing index and just add
    additional columns to it

    Return:
    A copy of df_in with updated MultiIndex. The index defines the groups of neurons, 
    i.e. neurons / fibers with the same index are assumed to be part of the same group.
    If replace=True, a new index is created; otherwise columns are added to any existing
    index.
    """
    vals = df_in[lst_props]
    vals.columns = [prefix + col for col in vals.columns]
    if not replace:
        vals = pandas.concat([df_in.index.to_frame(), vals], axis=1)

    df_in = df_in.set_index(pandas.MultiIndex.from_frame(vals)).sort_index()
    return df_in


def group_by_binned_properties(df_in, lst_props, bins, prefix="binned-", replace=True):
    """
    Defines a group of neurons by their binned values of a list of specified properties.
    That is, values are binned and then each unique combination bin-values defines a
    separate group. Values must be binnable with numpy.digitize!
    Groups are defined by the index of a DataFrame holding their properties.

    Input:
    df_in (pandas.DataFrame): Holds information on neurons or innervating fibers.
    One row per neuron / fiber, one column per property.
    lst_props (list-like): List of columns to use for grouping.
    bins: Defines how to bin the values. Can be an integer, defining the number
    of bins to use, or a list, defining the bin borders directly. Can also be a dict
    index by the name of the column that it applies to (if you want to use different
    bins for different columns).
    prefix (str): A prefix to preprepend to the names of the used columns/properties
    for their use in the index. This is to avoid having columns in the frame and index
    with the same name, which can break .groupby statements.
    replace (bool, default=True): Whether to replace the existing index and just add
    additional columns to it

    Return:
    A copy of df_in with updated MultiIndex. The index defines the groups of neurons, 
    i.e. neurons / fibers with the same index are assumed to be part of the same group.
    If replace=True, a new index is created; otherwise columns are added to any existing
    index.
    """
    def make_bins(values):
        if isinstance(bins, dict):
            b = bins[values.name]
        else:
            b = bins
        if not hasattr(b, "__iter__"):
            b = numpy.arange(numpy.min(values), numpy.max(values) + 1E-12, b)
        return b
    binned_vals = df_in[lst_props].apply(lambda x: numpy.digitize(x, bins=make_bins(x)), axis=0)
    binned_vals.columns = [prefix + col for col in binned_vals.columns]

    if not replace:
        binned_vals = pandas.concat([df_in.index.to_frame(), binned_vals], axis=1, copy=False)

    df_in = df_in.set_index(pandas.MultiIndex.from_frame(binned_vals)).sort_index()
    return df_in


def group_by_grid(df_in, lst_props, radius, shape="hexagonally", prefix="grid-", replace=True):
    """
    Defines a group of neurons by splitting their locations in a coordinate system
    into a regular grid.
    That is, the coordinate values are split into a grid and then each cell of the grid
    defines a separate group.
    Groups are defined by the index of a DataFrame holding their properties.

    Input:
    df_in (pandas.DataFrame): Holds information on neurons or innervating fibers.
    One row per neuron / fiber, one column per property.
    lst_props (list-like): List of columns holding the coordinates to use for grouping.
    Must have length 2, i.e. this only works for two-dimensional grids!
    radius (float): Defines the resolution of the grid. In the same units as the 
    coordinates given.
    shape (str, default: hexagonally): Defines the shape of the grid. Must one one of:
       - rhombically
       - triangularly
       - hexagonally
    prefix (str): A prefix to preprepend to the names of the used columns/properties
    for their use in the index. This is to avoid having columns in the frame and index
    with the same name, which can break .groupby statements.
    replace (bool, default=True): Whether to replace the existing index and just add
    additional columns to it

    Return:
    A copy of df_in with updated MultiIndex. The index defines the groups of neurons, 
    i.e. neurons / fibers with the same index are assumed to be part of the same group.
    If replace=True, a new index is created; otherwise columns are added to any existing
    index.
    """
    assert len(lst_props) == 2, "Need one column as x-coordinate and one as y-coordinate"
    tritilling = TriTille(radius)

    flat_locs = df_in[lst_props].rename(dict([(a, b) for a, b in zip(lst_props, ["x", "y"])]), axis=1)

    hexmap = getattr(tritilling, "bin_" + shape)(flat_locs, use_columns_row_indexing=False)
    grid = tritilling.locate_grid(hexmap)

    annotation = tritilling.annotate(grid, using_column_row=True)
    annotated_grid = grid.assign(subtarget=annotation.loc[grid.index])
    per_neuron_annotation = annotated_grid.loc[pandas.MultiIndex.from_frame(hexmap)]

    hexmap.columns = [prefix + col for col in hexmap.columns]
    per_neuron_annotation.columns = [prefix + col for col in per_neuron_annotation.columns]
    if not replace:
        hexmap = hexmap.reset_index()
    per_neuron_annotation = per_neuron_annotation.set_index(pandas.MultiIndex.from_frame(hexmap))

    df_in = df_in.set_index(per_neuron_annotation.index)
    df_in = pandas.concat([df_in, per_neuron_annotation], axis=1)
    return df_in


def flip(df_in, lst_values=None, index=GID, contract_values=False, categorical=False):
    """
    "Flips" the representation of a neuron group. I.e. instead of the index specifying the group,
    it will be indexed by one of the data columns (such as "gid") and the values will specify
    the group.
    Input:
    df_in: A pandas.DataFrame of neuron of fiber properties with groups defined by a MultiIndex.
    lst_values (list-like, optional): List of columns of the index of df_in to use as the new
    values of the output. If not provided, then all columns will be used.
    index (str, default: "gid"): Name of the column to use as the new index.
    contract_values (bool, default=False): If False, then the output will be a DataFrame with the
    columns specified by lst_values. If True, then it will be contracted to a Series by casting
    the columns to str and then concatenating them.
    categorical (bool, default=False): Only used in conjunction with contract_values=True. If True,
    then the values of the output will be a pandas.Categorical, which may be faster for some purposes.
    """
    if isinstance(df_in, pandas.DataFrame):
        df_in = df_in[index]
    idx_frame = df_in.index.to_frame()
    if lst_values is None:
        lst_values = idx_frame.columns.values
    idx_frame.index = df_in.values
    idx_frame = idx_frame[lst_values]

    if contract_values:
        new_name = ";".join(idx_frame.columns)
        idx_series = idx_frame.apply(lambda row: "_".join(map(str, row)), axis=1)
        idx_series.name = new_name
        if categorical:
            idx_series = pandas.Series(pandas.Categorical(idx_series), index=idx_series.index, name=idx_series.name)
        return idx_series
    return idx_frame


def count_overlap(df_one, df_two, in_column=GID):
    """
    Counts the amount of overlap in two neuron grouping schemes. For example: scheme A classifies according
    to mtype, scheme two according to etype. Then this yields the amount of neurons with a given mtype AND etype.
    That would be equivalent to grouping neurons according to [mtype, etype] and then doing .value_counts on the
    index, so I guess this is not that useful.
    """
    def prepare_overlap(df_pre):
        def execute_overlap(df_post):
            return len(numpy.intersect1d(df_pre[in_column], df_post[in_column]))

        return execute_overlap

    res = df_one.groupby(df_one.index.names).apply(
        lambda df_pre:
        df_two.groupby(df_two.index.names).apply(prepare_overlap(df_pre))
    )
    return res
